# Copyright information
__author__  = "Ole Weidner <oweidner@cct.lsu.edu>"
__date__    = "13 September 2009"
__version__ = "1.0"

# Import Ganga Threading Stuff
from GangaCore.Core.GangaThread.MTRunner import MTRunner, Data, Algorithm

# Setup Utility Logger
from GangaCore.Utility.logging import getLogger
logger = getLogger()

import os
import sys
import tarfile

# SAGA Job Package Imports
import saga.job

##############################################################################
##
class SAGAFileTransferTask:

    _attributes = ('jobObj')

    ##########################################################################
    ##
    def __init__(self, jobObj):
        self.jobObj  = jobObj

    ##########################################################################
    ##
    #def __eq__(self, other):
    #    """
    #    download task comparison based on job's FQID.
    #    """
    #    if self.jobObj.getFQID('.') == other.jobObj.getFQID('.'):
    #        return True
    #    else:
    #        return False

    ##########################################################################
    ##
    #def __str__(self):
    #    """
    #    represents the task by the job object
    #    """
    #    return 'downloading task for job %s' % self.jobObj.getFQID('.')


##############################################################################
##
class SAGAFileTransferAlgorithm(Algorithm):

    ##########################################################################
    ##
    def process(self, item):

        job = item.jobObj

        ## It is very likely that the job's downloading task has been created 
        ## and assigned in a previous monitoring loop ignore such kind of cases
        if job.status in ['completing', 'completed', 'failed']:
            return True

        job.updateStatus('completing')
        
        try:
            logger.info("poststaging output sandbox files")

            logger.info("  * %s", saga.url(job.backend.saga_job_out).url)
            stdout = saga.filesystem.file(saga.url(job.backend.saga_job_out));
            stdout.copy("file://localhost/"+job.getOutputWorkspace().getPath()+"stdout");
    
            logger.info("  * %s", saga.url(job.backend.saga_job_err).url)
            stderr = saga.filesystem.file(saga.url(job.backend.saga_job_err));
            stderr.copy("file://localhost/"+job.getOutputWorkspace().getPath()+"stderr");

            if len(job.outputsandbox) > 0 :
                output_sandbox = saga.url(job.backend.filesystem_url+"/"+job.backend.workdir_uuid+"/_output_sandbox.tgz")
                logger.info("  * %s", output_sandbox.url)
                osb = saga.filesystem.file(output_sandbox)
                osb.copy("file://localhost/"+job.getOutputWorkspace().getPath())
                
                ## Unpack the output sandbox and delete the archive
                osbpath = job.getOutputWorkspace().getPath()+"_output_sandbox.tgz"
                if os.path.exists(osbpath):
                    if os.system("tar -C %s -xzf %s"%(job.getOutputWorkspace().getPath(),job.getOutputWorkspace().getPath()+"/_output_sandbox.tgz")) != 0:
                        job.updateStatus('failed')
                        raise Exception('cannot upack output sandbox')
                    os.remove(osbpath)

            job.updateStatus('completed')
        
        except saga.exception as e:
            logger.error('exception caught while poststaging: %s', e.get_full_message())
            job.updateStatus('failed')
        
        return True

##############################################################################
##
class SAGAFileTransferManager(MTRunner):

    ##########################################################################
    ##
    def __init__(self, numThread=10):

        logger.debug('starting new MTRunner instance: %s.', "saga_file_transfer_manager")

        MTRunner.__init__(self, name='saga_file_transfer_manager', 
                          data=Data(collection=[]), 
                          algorithm=SAGAFileTransferAlgorithm())
        
        self.keepAlive = True
        self.numThread = numThread

    ##########################################################################
    ##
    def addTask(self, job):
        
        task = SAGAFileTransferTask(job)
        self.addDataItem(task)
        
        logger.debug( 'added new file transfer thread for job id %s', job.id )
        
        return True
        
        




from GangaCore.Core.exceptions import ApplicationConfigurationError
from GangaCore.GPIDev.Base import GangaObject
from GangaCore.GPIDev.Schema import *
from GangaCore.GPIDev.Lib.File import *

from GangaCore.Lib.LCG import LCGJobConfig
from GangaCore.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers
from GangaCore.Utility.Config import makeConfig, ConfigError
from GangaCore.Utility.logging import getLogger
from GangaCore.Utility.files import expandfilename

from GangaCore.GPIDev.Adapters.IApplication import IApplication
from GangaCore.GPIDev.Adapters.IRuntimeHandler import IRuntimeHandler
from GangaCore.GPIDev.Adapters.StandardJobConfig import StandardJobConfig

from GangaCore.Lib.Mergers.Merger import *

# ------------------------------------------------------
# Spider RTHandler
def GetConfig(app):
    """prepare the subjob specific configuration"""
    
    #print "Preparing camont spider job"
    import time
    job = app._getParent()
    
    # set up the executable script
    exe = os.path.join(os.path.dirname(__file__),'web_spider.py')

    # Create the input list from the selected domains
    queue_lists = {}
    max_queue = 0     # current maximum number of links per domain

    for dom in app.domains:

        # Check the domain repository is there
        dir =  os.path.join( app.repository_loc, dom )
        if not os.path.exists( dir ):
            os.system("mkdir -p " + dir)
            open( os.path.join(dir, app._queuelistfilename ), "w" ).write( "http://" + dom + "/\n" )

        # load up the queue lists for each of the domains and find the maximum
        queue_file = open( os.path.join(dir, app._queuelistfilename ) )

        list = queue_file.readlines()

        if len(list) > max_queue:
            max_queue = len(list)

        queue_lists[dom] = list
        queue_file.close()

    # check for daily limit
    if max_queue > app.max_links:
        max_queue = app.max_links

    # loop and fill the input list
    q = 0
    input_list = []
    while q < max_queue:

        for dom in app.domains:

            list = queue_lists[dom]
            if q  < len( list ):
                input_list.append(queue_lists[dom][q])
            else:
                input_list.append('SLEEP\n')

        q = q + 1

    # remove these links from the queue list
    for dom in app.domains:
        
        dir =  os.path.join( app.repository_loc, dom )
        queue_file = open( os.path.join(dir, app._queuelistfilename ), "w" )
        list = queue_lists[dom]
        
        if len(list) <= max_queue:
            queue_file.close()
        else:
            for ln in list[max_queue:]:
                queue_file.write(ln)
            queue_file.close()

    # Create the input list file
    inputfile = open( app._inputlistfilename, "w")
    if len(input_list) == 0:
        raise ApplicationConfigurationError(None,'No input links found')

    for ln in input_list:
        inputfile.write(ln)

    inputfile.close()

    # create the extension list
    ext_file = open(app._extensionfilename, "w")
    for e in app.image_ext:
        ext_file.write(e + '\n')

    ext_file.write("--\n")
    for e in app.file_ext:
        ext_file.write(e + '\n')

    ext_file.close()

    # create the safe domains list
    safedoms_file = open(app._safedomsfilename, "w")
    for sd in app.safe_domains:
        safedoms_file.write(sd + '\n')

    safedoms_file.close()

    # set the environment
    environment = {'SPIDER_PAYLOAD' : os.path.basename( app.payload.name ) }

    # generate the config
    inputbox = [ File(os.path.join(os.path.dirname(__file__),'BeautifulSoup.py')),
                 File(app._inputlistfilename),
                 File(app._extensionfilename),
                 File(app._safedomsfilename),
                 File(app.payload.name )]
    inputbox += job.inputsandbox
    
    outputbox = [ 'image_info.txt', 'ext_info.txt', 'ext_list.txt', 'img_list.txt', 'links_list.txt', 'viewed_list.txt', 'dom_list.txt' ]

    outputbox.append( app.payload_output )

    return exe, inputbox, outputbox, environment

class SpiderLCGRTHandler(IRuntimeHandler):
    """Spider LCG RT Handler"""

    def prepare(self,app,appsubconfig,appmasterconfig,jobmasterconfig):
        """prepare the subjob specific configuration"""

        exe, inputsandbox, outputsandbox, env = GetConfig(app)
        lcg_config = LCGJobConfig(File(exe), inputsandbox, [], outputsandbox, env, [], None)
        
        return lcg_config

class SpiderLocalRTHandler(IRuntimeHandler):
    """Spider LCG RT Handler"""

    def prepare(self,app,appsubconfig,appmasterconfig,jobmasterconfig):
        """prepare the subjob specific configuration"""

        exe, inputsandbox, outputsandbox, env = GetConfig(app)
        local_config = StandardJobConfig(File(exe), inputsandbox, [], outputsandbox, env)
        return local_config


allHandlers.add('Spider','LCG', SpiderLCGRTHandler)
allHandlers.add('Spider','PBS', SpiderLocalRTHandler)
allHandlers.add('Spider','Local', SpiderLocalRTHandler)



from __future__ import absolute_import
##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: Proxy.py,v 1.2.4.3 2009-07-10 11:29:27 ebke Exp $
##########################################################################

import GangaCore.Utility.logging
from GangaCore.Utility.Config import getConfig

from GangaCore.GPIDev.Schema import ComponentItem

from GangaCore.Core.exceptions import GangaException, GangaValueError
from GangaCore.GPIDev.Base.Objects import Node, GangaObject, ObjectMetaclass, _getName
from GangaCore.Core.exceptions import GangaAttributeError, ProtectedAttributeError, ReadOnlyObjectError, TypeMismatchError

import collections
import functools
import os

from inspect import isclass, getargspec

import types

from copy import deepcopy

implRef = '_impl'
proxyClass = '_proxyClass'
proxyObject = '_proxyObject'

prepconfig = getConfig('Preparable')

logger = GangaCore.Utility.logging.getLogger()

# some proxy related convieniance methods

_knownLists = None

_stored_Interface = None

_eval_cache = {}

def setProxyInterface(my_interface):
    """ Set the proxy interface, not strictly needed for GangaCore.GPI but good practice as we move to 'ganga' """
    global _stored_Interface
    _stored_Interface = my_interface

def getProxyInterface():
    """ Get the proxy interface  GangaCore.GPI by default, 'ganga' if it's be set to this """
    if not _stored_Interface:
        import GangaCore.GPI
        setProxyInterface(GangaCore.GPI)
    return _stored_Interface

def getRuntimeGPIObject(obj_name, silent=False, evalClass=True):
    """ Get, or attempt to get an object from the GPI, if it exists then return a new instance if a class or an object if it's not
       If it doesn't exist attempt to evaluate the obj_name as a string like a standard python object
       If it's none of the above then return 'None' rather than the object string which was input
       Args:
           obj_name(str): This is the object we want to get from the GPI
           silent(bool): Should we be silent about errors?
           evalClass(bool): Should we create an instance of a class object when it's returned"""
    interface = getProxyInterface()
    if obj_name in interface.__dict__:
        this_obj = interface.__dict__[obj_name]
        if evalClass and isclass(this_obj):
            return this_obj()
        else:
            return this_obj
    else:
        returnable = raw_eval(obj_name)
        if returnable == obj_name:
            if silent is False:
                logger.error("Cannot find Object: '%s' in GPI. Returning None." % obj_name)
            return None
        return returnable

def runtimeEvalString(this_obj, attr_name, val):
    """
     Return the evaluated value of the 'val' after checking the schema and attributes associated with this_obj and attr_name
     If the attribute or the Schema are or allow for string objects then val is not evaluated but if it does allow for non string objects and isn't then an eval is performed
     This is ugly and is a direct consequence of allowing j.backend = 'Dirac' which in this authors (rcurrie) opinion is going to hurt us later
    """

    ## Don't check or try to auto-eval non-string objects
    if not isinstance(val, str):
        return val

    raw_obj = stripProxy(this_obj)
    shouldEval = None

    ## Lets check the Schema to see if a string object is allowed here
    ## If this is a ComponentItem we know in advance that we need to try and evaluate this
    ## If the object is NOT a ComponentItem but is still in the schema, check to see if the object is allowed to be a string or not
    ## If an object is NOT allowed to be a string then it should be 'eval'-ed
    if hasattr(raw_obj, '_schema'):

        if not raw_obj._schema.hasAttribute(attr_name):
            shouldEval = True
        else:
            this_attr = raw_obj._schema.getItem(attr_name)
            if isType(this_attr, ComponentItem):
                ## This is a component Item and isn't equivalent to a string
                shouldEval = True
            else:
                allowedTypes = this_attr.getProperties()['typelist']
                for this_type in allowedTypes:
                    if this_type == str:
                        ## This type is a string and shouldn't be evaluated
                        shouldEval = False
                        break
                    else:
                        ## This type is written as a string so need to work out what it is
                        if type(this_type) == str:
                            try:
                                interface = getProxyInterface() 
                                eval_type = eval(this_type, interface.__dict__)
                                if eval_type == str:
                                    ## This type was written as "str" ... slightly annoying but OK...
                                    shouldEval = False
                                    break
                                else:
                                    ## This type is NOT a string so based on this we should Eval
                                    shouldEval = True
                            except Exception as err:
                                logger.debug("Failed to evalute type: %s" % this_type)
                                logger.debug("Err: %s" % err)
                                ## We can't eval in this case. It may just be the type which is broken
                                shouldEval = True
                        else:
                            ## This type isn't in a string format so we don't need to work out what it is
                            shouldEval = True

    ## If the attribute is not in the Schema lets see if this class instance knows about this object or not
    ## If the attribute is NOT a string but is in this instane then we should try and eval
    ## But if it's known by the instance and is a string, we should just use the value
    elif hasattr(raw_obj, attr_name):

        if type(getattr(raw_obj, attr_name)) == str:
            shouldEval = False
        else:
            shouldEval = True

    ## If the object is not in the schema then try and eval the object anyway as we a-priori don't know any better
    ## THIS IS POTENTIALLY DANGEROUS AND IF A LOT OF USERS COMPLAIN THIS SHOULD BE REVERSED TO FALSE!!!
    else:
        shouldEval = True

    assert(shouldEval is not None)

    if shouldEval is True:
        new_val = raw_eval(val)
    else:
        new_val = val

    return new_val



def raw_eval(val):
    """
     Attempts to evaluate the val object and return the object it evaluates to if it is a Python object
     Makes use of basic caching as we don't expect that things at this level should change.
     Args:
        val(str): This is the string which we're looking to evaluate from 'in front' of the proxy
     """

    if val in _eval_cache:
        return deepcopy(_eval_cache[val])

    try:
        interface = getProxyInterface() 
        temp_val = eval(val, interface.__dict__)
        if isclass(temp_val):
            new_val = temp_val()
        else:
            new_val = temp_val
    except Exception as err:
        ## Useful for debugging these
        ## import traceback; traceback.print_stack()
        logger.debug("Proxy Cannot evaluate v=: '%s'" % val)
        logger.debug("Using raw value instead")
        new_val = val

    raw_val = stripProxy(new_val)
    if hasattr(raw_val, '_auto__init__'):
        raw_val._auto__init__()

    _eval_cache[val] = new_val

    return deepcopy(new_val)

def getKnownLists():
    """ Returns the list of iterable objects, tuple, list and maybe GangaList which we can use here due to import """
    global _knownLists
    if _knownLists is None:
        try:
            from GangaCore.GPIDev.Lib.GangaList.GangaList import GangaList
        except ImportError:
            _knownLists = None
            return (tuple, list)
        _knownLists = (tuple, list, GangaList)
    return _knownLists

def isProxy(obj):
    """Checks if an object is a proxy
    Args:
        obj (object): This may be an instance or a class
    """
#    return isinstance(obj.__class__, GPIProxyObject)
    # Alex changed for below as doesn't check class objects, only instances
    # e.g. isProxy(DiracFile) fails at the Ganga prompt
    if isclass(obj):
        return issubclass(obj, GPIProxyObject) or hasattr(obj, implRef)
    else:
        obj_class = obj.__class__
        return issubclass(obj_class, GPIProxyObject) or hasattr(obj_class, implRef)

def isType(_obj, type_or_seq):
    """Checks whether on object is of the specified type, stripping proxies as needed.
    Args:
        obj (object): This may be an instance or a class
        type_or_seq (type, list, tuple, GangaList): This may be an individual type or an iterable list of types
    """

    obj = stripProxy(_obj)

    bare_type_or_seq = stripProxy(type_or_seq)

    ## Here to avoid circular GangaObject dependencies
    ## is type_or_seq iterable?
    if isinstance(type_or_seq, getKnownLists()):
        clean_list = []
        for type_obj in type_or_seq:
            if type_obj != str and type(type_obj) != type(str) and (not isclass(type_obj)):
                clean_list.append(type(stripProxy(type_obj)))
            elif isclass(type_obj):
                clean_list.append(type_obj)
            else:
                clean_list.append(type_obj)

        return isinstance(obj, tuple(clean_list))

    else:
        return isinstance(obj, bare_type_or_seq)

def getName(_obj):
    """Strip any proxy and then return an objects name
    Args:
        _obj (object): This may be an instance or a class"""
    obj = stripProxy(_obj)
    returnable = _getName(obj)
    return returnable

def is_namedtuple_instance(obj):
    if len(type(obj).__bases__) != 1 or not isinstance(obj, tuple):
        return False
    fields = getattr(obj, '_fields', None)
    if not isinstance(fields, tuple):
        return False
    return all(isinstance(n, str) for n in fields)

def stripProxy(obj):
    """Removes the proxy if there is one
    Args:
        obj (object): This may be an instance or a class
    """
    if is_namedtuple_instance(obj):
        return type(obj)(*[stripProxy(_) for _ in obj])
    elif isinstance(obj, (list, tuple)):
        return type(obj)(stripProxy(_) for _ in obj)
    elif isinstance(obj, dict):
        return dict((k, stripProxy(v)) for k, v in obj.items())
    elif hasattr(obj, implRef):
        return getattr(obj, implRef)
    else:
        return obj


def addProxy(obj):
    """Adds a proxy to a GangaObject instance or class
    Args:
        obj (GangaObject): This may be a Ganga object which you're wanting to add a proxy to
    """
    if isType(obj, GangaObject):
        if not isProxy(obj):
            if hasattr(obj, proxyObject):
                return getattr(obj, proxyObject)
            else:
                return GPIProxyObjectFactory(obj)
    elif isclass(obj) and issubclass(obj, GangaObject):
        return getProxyClass(obj)
    elif is_namedtuple_instance(obj):
        return type(obj)(*[addProxy(_) for _ in obj])
    elif isinstance(obj, (list, tuple)):
        return type(obj)(addProxy(_) for _ in obj)
    elif isinstance(obj, dict):
        return dict((k, addProxy(v)) for k, v in obj.items())
    return obj


def getProxyAttr(obj, attr_name):
    """Gets an attribute from a proxied object"""
    return getattr(stripProxy(obj), attr_name)


def runProxyMethod(obj, method_name, *args):
    """Calls a method on the object, removing the proxy if needed"""
    fp = getProxyAttr(obj, method_name)
    return fp(*args)


def export(method):
    """
    Decorate a GangaObject method to be exported to the GPI
    """
    method.exported = True
    return method

# apply object conversion or if it fails, strip the proxy and extract the
# object implementation


def stripComponentObject(v, cfilter, item):

    def getImpl(v):
        if v is None:
            if not item['optional']:
                raise TypeMismatchError(None, 'component(%s) is mandatory and None may not be used' % getName(item))
                return v
            else:
                return None
        if isType(v, GangaObject):
            return v
        if not isinstance(v, GPIProxyObject):
            raise TypeMismatchError("cannot assign value '%s', expected a '%s' object " % (repr(v), item['category']))
        return stripProxy(v)

    vv = cfilter(v, item)
    if vv is None:
        return getImpl(v)
    else:
        return vv

from GangaCore.GPIDev.TypeCheck import _valueTypeAllowed
valueTypeAllowed = lambda val, valTypeList: _valueTypeAllowed(val, valTypeList, logger)


class ProxyDataDescriptor(object):

    __slots__ = ('_name', )

    def __init__(self, name):
        """
        Descriptor which sits in fromnt  of raw unproxied objects
        Args:
            name (str): Name of the attribute which we're looking after here
        """
        self._name = name

    # apply object conversion or if it failes, make the wrapper proxy
    def disguiseComponentObject(self, v):
        # get the proxy for implementation object
        def getProxy(v):
            if not isType(v, GangaObject):
                raise GangaAttributeError("invalid type: cannot assign '%s' to attribute '%s'" % (repr(v), getName(self)))
            return GPIProxyObjectFactory(v)

        # convert implementation object to GPI value according to the
        # static method defined in the implementation object
        def object2value(v):
            return None

        vv = object2value(v)
        if vv is None:
            # allow None to be a legal value for optional component items
            if v is None:
                return None
            else:
                return getProxy(v)

        else:
            return vv

    # apply attribute conversion
    def disguiseAttribute(self, v):
        if isType(v, GangaObject):
            return GPIProxyObjectFactory(v)
        return v

    def __get__(self, obj, cls):

        # at class level return a helper object (for textual description)
        if obj is None:
            # return Schema.make_helper(getattr(getattr(cls, implRef), getName(self)))
            return getattr(stripProxy(cls), getName(self))

        raw_obj = stripProxy(obj)
        try:
            val = getattr(raw_obj, getName(self))
        except Exception as err:
            if getName(self) in raw_obj.__dict__:
                val = raw_obj.__dict__[getName(self)]
            else:
                val = getattr(raw_obj, getName(self))

        # wrap proxy
        item = raw_obj._schema[getName(self)]

        if item['proxy_get']:
            return getattr(raw_obj, item['proxy_get'])()

        if isType(item, ComponentItem):
            disguiser = self.disguiseComponentObject
        else:
            disguiser = self.disguiseAttribute

        ## FIXME Add GangaList?
        if item['sequence'] and isType(val, list):
            from GangaCore.GPIDev.Lib.GangaList.GangaList import makeGangaList
            val = makeGangaList(val, disguiser)

        returnable = disguiser(val)
        

        if isType(returnable, GangaObject):
            return addProxy(returnable)
        else:
            return returnable

    @staticmethod
    def _check_type(obj, val, attr_name):
        item = stripProxy(obj)._schema[attr_name]
        return item._check_type(val, attr_name)

    # apply attribute conversion
    @staticmethod
    def _stripAttribute(obj, v, name):
        # just warn
        # print '**** checking',v,v.__class__,
        # isinstance(val,GPIProxyObject)
        new_v = None
        if isinstance(v, list):
            from GangaCore.GPIDev.Lib.GangaList.GangaList import GangaList
            v_new = GangaList()
            for elem in v:
                v_new.append(elem)
        if isinstance(v, GPIProxyObject) or hasattr(v, implRef):
            new_v = stripProxy(v)
            logger.debug('%s property: assigned a component object (%s used)' % (name, implRef))

        if new_v is None:
            new_v = v
        return obj._attribute_filter__set__(name, new_v)

    @staticmethod
    def __app_set__(obj, val):

        if not hasattr(obj, 'application') or obj.application is None:
            return

        if hasattr(obj.application, '_is_prepared'):

            #a=Job(); a.prepare(); a.application=Executable()
            if obj.application.is_prepared not in [None, True] and\
                 hasattr(val, 'is_prepared') and val.is_prepared is None:
                 logger.debug('Overwriting a prepared application with one that is unprepared')
                 obj.application.unprepare()

            #a=Job(); b=Executable(); b.prepare(); a.application=b
        elif obj.application.is_prepared is not True:
            if hasattr(val, 'is_prepared'):
                if val.is_prepared not in [None, True]:
                    from GangaCore.Core.GangaRepository import getRegistry
                    shareref = GPIProxyObjectFactory(getRegistry("prep").getShareRef())
                    logger.debug('Overwriting application with a prepared one')
                    raw_app = stripProxy(obj.application)
                    if raw_app != val:
                        raw_app.unprepare()
                        shareref.increase(val.is_prepared)

            # check that the shared directory actually exists before
            # assigning the (prepared) application to a job
            if hasattr(val, 'is_prepared'):
                if val.is_prepared not in [None, True]:
                    if hasattr(val.is_prepared, 'name'):
                        from GangaCore.Utility.files import expandfilename
                        Config_conf = getConfig('Configuration')
                        shared_path = os.path.join(expandfilename(Config_conf['gangadir']), 'shared', Config_conf['user'])
                        if not os.path.isdir(os.path.join(shared_path, val.is_prepared.name)):
                            logger.error('ShareDir directory not found: %s' % val.is_prepared.name)

    @staticmethod
    def __prep_set__(obj, val):

        # if we set is_prepared to None in the GPI, that should effectively
        # unprepare the application
        if val is None:
            if obj.is_prepared is not None:
                logger.info('Unpreparing application.')
                obj.unprepare()

        # Replace is_prepared on an application for another ShareDir object
        if hasattr(obj, '_getRegistry'):
            from GangaCore.GPIDev.Lib.File import ShareDir
            if obj._getRegistry() is not None and isType(val, ShareDir):
                logger.debug('Overwriting is_prepared attribute with a ShareDir object')
                # it's safe to unprepare 'not-prepared' applications.
                obj.unprepare()
                from GangaCore.Core.GangaRepository import getRegistry
                shareref = getRegistry("prep").getShareRef()
                shareref.increase(val)

        if isinstance(val, str):
            logger.error("Setting string type to 'is_prepared'")
            #import traceback
            #traceback.print_stack()

    @staticmethod
    def __sequence_set__(stripper, obj, val, name):
        item = obj._schema[name]
        # we need to explicitly check for the list type, because simple
        # values (such as strings) may be iterable
        new_v = None
        from GangaCore.GPIDev.Lib.GangaList.GangaList import makeGangaList
        if isType(val, getKnownLists()):

            # create GangaList
            if stripper is not None:
                new_v = makeGangaList(val, stripper)
            else:
                temp_v = ProxyDataDescriptor._stripAttribute(obj, val, name)
                new_v = makeGangaList(temp_v)
        else:
            # val is not iterable
            if item['strict_sequence']:
                raise GangaAttributeError('cannot assign a simple value %s to a strict sequence attribute %s.%s (a list is expected instead)' % (repr(val), getName(obj), name))
            if stripper is not None:
                new_v = makeGangaList(stripper(val))
            else:
                temp_v = ProxyDataDescriptor._stripAttribute(obj, val, name)
                new_v = makeGangaList(temp_v)

        if new_v is None:
            new_v = val
        return new_v

    @staticmethod
    def __preparable_set__(obj, val, name):
        if obj.is_prepared is not None:
            if obj.is_prepared is not True:
                raise ProtectedAttributeError('AttributeError: "%s" attribute belongs to a prepared application and so cannot be modified.\
                                                unprepare() the application or copy the job/application (using j.copy(unprepare=True)) and modify that new instance.' % (name,))

    ## Inspect this given item to determine if it has editable attributes if it has been set as read-only
    ## Curently Unused although may be useful to keep
    @staticmethod
    def __subitems_read_only(obj):
        can_be_modified = []
        for name, item in obj._schema.allItems():
            ## This object inherits from Node therefore likely has a schema too.
            obj_attr = getattr(obj, name)
            if isType(obj_attr, Node):
                can_be_modified.append( ProxyDataDescriptor.__subitems_read_only(obj_attr) )
            else:

                ## This object doesn't inherit from Node and therefore needs to be evaluated
                if item.getProperties()['changable_at_resubmit']:
                    can_be_modified.append( True )
                else:
                    can_be_modified.append( False )
        
        can_modify = False
        for i in can_be_modified:
            can_modify = can_modify or i

        return can_modify

    @staticmethod
    def __recursive_strip(_val):
        ## Strip the proxies recursively for things like nested lists
        raw_val = stripProxy(_val)
        if isinstance(_val, collections.Sequence) and not isinstance(_val, basestring):
            val = raw_val.__class__()
            if isinstance(val, dict):
                for _key, elem in _val.iteritems():
                    if isType(_key, GangaObject):
                        key = stripProxy(_key)
                    else:
                        key = _key
                    if isType(elem, GangaObject):
                        val[key] = ProxyDataDescriptor.__recursive_strip(stripProxy(elem))
                    else:
                        val[key] = elem
            else:
                for elem in _val:
                    if isType(elem, GangaObject):
                        val.append(ProxyDataDescriptor.__recursive_strip(stripProxy(elem)))
                    else:
                        val.append(elem)
        else:
            val = raw_val
        return val

    @staticmethod
    def _process_set_value(raw_obj, _val, attr_name, check_read_only=True):
        """
        Process an incoming attribute value.

        Args:fdef _init
            raw_obj: the object the value is being assigned to
            _val: the value being assigned
            attr_name: the name of the attribute being assigned
            check_read_only: enforce the checks of read-only objects. This makes sense to be disabled during object construction.

        Returns:
            The processed value
        """

        val = ProxyDataDescriptor.__recursive_strip(_val)

        new_val = None

        if not raw_obj._schema.hasAttribute(attr_name):
            raise GangaAttributeError("Cannot assign %s, as it is NOT an attribute in the schema for class: %s" % (attr_name, getName(obj))) 

        #logger.debug("__set__")
        item = raw_obj._schema[attr_name]
        if item['protected']:
            raise ProtectedAttributeError('"%s" attribute is protected and cannot be modified' % (attr_name,))
        if raw_obj._readonly():

            if not item.getProperties()['changable_at_resubmit']:
                raise ReadOnlyObjectError('object %s is read-only and attribute "%s" cannot be modified now' % (repr(addProxy(raw_obj)), attr_name))

        if check_read_only:
            # mechanism for locking of preparable attributes
            if item['preparable']:
                ## Does not modify val
                ProxyDataDescriptor.__preparable_set__(raw_obj, val, attr_name)

        # if we set is_prepared to None in the GPI, that should effectively
        # unprepare the application
        if attr_name == 'is_prepared':
            # Replace is_prepared on an application for another ShareDir object
            ## Does not modify val
            ProxyDataDescriptor.__prep_set__(raw_obj, val)

        # catch assignment of 'something'  to a preparable application
        if attr_name == 'application':
            ## Does not modify val
            ProxyDataDescriptor.__app_set__(raw_obj, val)

        # unwrap proxy
        if item.isA(ComponentItem):
            from .Filters import allComponentFilters
            cfilter = allComponentFilters[item['category']]
            stripper = lambda v: stripComponentObject(v, cfilter, item)
        else:
            stripper = None

        if item['sequence']:
            ## Does not modify val
            new_val = ProxyDataDescriptor.__sequence_set__(stripper, raw_obj, val, attr_name)
        else:
            if stripper is not None:
                ## Shouldn't modify val
                new_val = stripper(val)
            else:
                ## Does not modify val
                new_val = ProxyDataDescriptor._stripAttribute(raw_obj, val, attr_name)

        if new_val is None and val is not None:
            new_val = val

        final_val = None
        # apply attribute filter to component items
        if item.isA(ComponentItem):
            ## Does not modify val
            final_val = ProxyDataDescriptor._stripAttribute(raw_obj, new_val, attr_name)
        else:
            final_val = new_val

        if final_val is None and val is not None:
            final_val = val

        ## Does not modify val?
        ProxyDataDescriptor._check_type(raw_obj, final_val, attr_name)

        return final_val

    def __set__(self, obj, _val):
        # self is the attribute we're about to change
        # obj is the object we're about to make the change in
        # val is the value we're setting the attribute to.
        # item is the schema entry of the attribute we're about to change

        attr_name = getName(self)

        raw_obj = stripProxy(obj)

        final_val = ProxyDataDescriptor._process_set_value(raw_obj, _val, attr_name)

        setattr(raw_obj, attr_name, final_val)


def proxy_wrap(f):
    # type: (Callable) -> Callable
    """
    A decorator to strip the proxy from all incoming arguments
    (including ``self`` if it's a method) and add one to the return
    value.
    """

    @functools.wraps(f)
    def proxy_wrapped(*args, **kwargs):
        s_args = [stripProxy(a) for a in args]
        s_kwargs = dict((name, stripProxy(a)) for name, a in kwargs.items())
        try:
            r = f(*s_args, **s_kwargs)
        except KeyboardInterrupt:
            logger.error("Command was interrupted by a Ctrl+C event!")
            logger.error("This can lead to inconsistencies")
            return
        return addProxy(r)

    return proxy_wrapped


class ProxyMethodDescriptor(object):

    __slots__ = ('_name', '_internal_name', '__doc__')

    def __init__(self, name, internal_name):
        self._name = name
        self._internal_name = internal_name

    def __get__(self, obj, cls):
        try:
            if obj is None:
                method = getattr(stripProxy(cls), self._internal_name)
            else:
                method = getattr(stripProxy(obj), self._internal_name)
            return proxy_wrap(method)
        except Exception as err:
            logger.error("%s" % err)
            raise

##########################################################################

# helper to create a wrapper for an existing ganga object

_proxyClassDict={}

def addProxyClass(some_class):
    ## CANNOT USE THE ._name (hence getName) HERE DUE TO REQUIREMENTS OF THE OBJECT IN GPI BEING SANE!!!
    class_name = some_class.__name__
    if class_name not in _proxyClassDict:
        _proxyClassDict[class_name] = GPIProxyClassFactory(class_name, some_class)    
    setattr(some_class, proxyClass, _proxyClassDict[class_name])

def getProxyClass(some_class):
    class_name = some_class.__name__
    if not isclass(some_class):
        raise GangaException("Cannot perform getProxyClass on a non-class object: %s:: %s" % (class_name, some_class))
    if not issubclass(some_class, GangaObject):
        raise GangaException("Cannot perform getProxyClass on class which is not a subclass of GangaObject: %s:: %s" % (class_name, some_class))
    proxy_class = getattr(some_class, proxyClass, None)
    ## It's possible we ourselves have added a proxy to the base class which we're now inheriting here.
    ## To avoid giving a proxy from Dataset to LHCbDataset and equivalent we'll check against our list of already-found class names.
    if proxy_class is None or class_name not in _proxyClassDict:
        addProxyClass(some_class)
        proxy_class = getattr(some_class, proxyClass)
    return proxy_class

def GPIProxyObjectFactory(_obj):
    # type: (GangaObject) -> GPIProxyObject
    """
    This function _must_ be passed a raw GangaObject. Use :function:`addProxy` for a safe version

    Args:
        _obj (GangaObject): the object to wrap

    Returns:
        a proxy object
    """
    if hasattr(_obj, proxyObject):
        return getattr(_obj, proxyObject)
    if not isType(_obj, GangaObject):
        raise GangaException("%s is NOT a Proxyable object" % type(_obj))

    obj_class = _obj.__class__

    proxy_class = getProxyClass(obj_class)

    proxy_object = proxy_class(_proxy_impl_obj_to_wrap=_obj)

    setattr(_obj, proxyObject, proxy_object)

    return proxy_object

# this class serves only as a 'tag' for all generated GPI proxy classes
# so we can test with isinstance rather then relying on more generic but
# less user friendly checking of attribute x._impl


class GPIProxyObject(object):
    __slots__ = list()
    pass

# create a new GPI class for a given ganga (plugin) class


def GPIProxyClassFactory(name, pluginclass):
    # type: (str, type(GangaObject)) -> type(GPIProxyObject)
    """
    Args:
        name: the name of the proxy class
        pluginclass: the ``GangaObject`` subclass to wrap

    Returns:
        a new type which wraps ``pluginclass``
    """

    def helptext(f, s):
        if name == '' or name is None:
            _name = ' '
        else:
            _name = name
        f.__doc__ = s % {'classname': _name, 'objname': _name.lower(), 'shortvarname': _name[0].lower()}

    # construct the class on-the-fly using the functions below as methods for
    # the new class

    def _init(self, *args, **kwds):

        ## Zero-th fully initialize self before moving on
        GPIProxyObject.__init__(self)

        ## THE ORDER IN HOW AN OBJECT IS INITIALIZED IS IMPORTANT AND HAS BEEN DOUBLE CHECKED - rcurrie


        ## If we're only constructing a raw Proxy to wrap an existing object lets wrap that and return
        proxy_obj_str = '_proxy_impl_obj_to_wrap'

        if proxy_obj_str in kwds:
            instance = kwds[proxy_obj_str]
            ## Even if we're wrapping something such as here make sure we set all of the proxy related attributes correctly.
            ## Setting of these attributes shold be done here within this class and should probably be properly be done on proxy construction. aka. here
        else:
            ## FIRST INITALIZE A RAW OBJECT INSTANCE CORRESPONDING TO 'pluginclass'
            ## Object was not passed by construction so need to construct new object for internal use
            # Case 1 j = Job(myExistingJob)            # We want to perform a deepcopy
            arg_len = len(args)
            if arg_len == 1 and isinstance(args[0], pluginclass):
                instance = deepcopy(stripProxy(args[0]))
            # Case 2 file_ = LocalFile('myFile.txt')   # We need to pass the (stripped) arguments to the constructor only if the 
            # Remember self = 1
            # For the moment we're warning the user until it's clear this is a safe thing to do, aka once all classes are deemed safe
            # The args will simply be passed through regardless
            elif arg_len == 0:
                instance = pluginclass.getNew(should_init=True)
            elif arg_len < len(getargspec(pluginclass.__init__)[0]):
                clean_args = (stripProxy(arg) for arg in args)
                instance = pluginclass(*clean_args)
            else:
                # In the future we will just pass the args to the classes directly and throw excepions, but for now we're trying to maintain old behavior
                logger.warning("Cannot use arguments: '%s' for constructing class type '%s'. Ignoring." % (args, getName(pluginclass)))
                logger.warning("Please contact the Ganga developers if you believe this is an error!")
                instance = pluginclass()

        ## Avoid intercepting any of the setter method associated with the implRef as they could trigger loading from disk
        ## These are protected objects in the setter and it will throw an exception if they're altered
        setattr(self, implRef, instance)
        instance.__dict__[proxyObject] = self

        ## Need to avoid any setter methods for GangaObjects
        ## Would be very nice to remove this entirely as I'm not sure a GangaObject should worry about it's proxy (if any)


        ## SECOND WE NEED TO MAKE SURE THAT OBJECT ID IS CORRECT AND THIS DOES THINGS LIKE REGISTER A JOB WITH THE REPO

        if proxy_obj_str in kwds:
            #instance._auto__init__()
            # wrapping not constructing so can exit after determining that the proxy attributes are setup correctly
            return

        from GangaCore.GPIDev.Base.Objects import do_not_copy
        ## All objects with an _auto__init__ method need to have that method called and we set the various node attributes here based upon the schema
        for key, _val in instance._schema.allItems():
            if not _val['getter'] and key not in instance._data:
                val = instance._schema.getDefaultValue(key)
                if isinstance(val, GangaObject):
                    val._auto__init__()
                instance.setSchemaAttribute(key, instance._attribute_filter__set__(key, val))

        instance._auto__init__()

        ## THIRD ALLOW FOR APPLICATION AND IS_PREPARED etc TO TRIGGER RELAVENT CODE AND SET THE KEYWORDS FROM THE SCHEMA AGAIN
        ## THIS IS MAINLY FOR THE FIRST EXAMPLE ABOVE

        ## THIS CORRECTLY APPLIES A PROXY TO ALL OBJECT ATTRIBUTES OF AN OBJECT CREATED WITHIN THE GPI

        # initialize all properties from keywords of the constructor
        for k in kwds:
            if instance._schema.hasAttribute(k):
                # This calls the same logic when assigning a named attribute as when we're assigning it to the object
                # There is logic here which we 'could' duplicate but it is over 100 lines of code which then is duplicating funtionality written elsewhere
                try:
                    val = ProxyDataDescriptor._process_set_value(instance, kwds[k], k, False)
                except Exception as err:
                    logger.warning('Error assigning following value to attribute: \'%s\'' % k)
                    logger.warning('value: \'%s\'' % str(kwds[k]))
                    logger.warning('Error: \'%s\'' % str(err))
                    raise GangaValueError('Error constructing object of type: \'%s\'' % getName(instance))
                if isinstance(val, GangaObject):
                    val._auto__init__()
                setattr(instance, k, val)
            else:
                logger.warning('keyword argument in the %s constructor ignored: %s=%s (not defined in the schema)', name, k, kwds[k])

        ## end of _init
        return

    from GangaCore.Utility.strings import ItemizedTextParagraph

    itbuf = ItemizedTextParagraph('Properties:', 80, ' ', '')

    for n, item in pluginclass._schema.allItems():
        if not item['hidden']:
            itbuf.addLine(n, item.describe())

    if not pluginclass.__doc__:
        pluginclass.__doc__ = 'Documentation missing.'

    pluginclass.__doc__.strip()
    pluginclass.__doc__ += "\n\n"

    publicdoc = pluginclass.__doc__ + itbuf.getString()

    helptext(_init, """GPI %(classname)s object constructor:
    %(classname)s() : create %(objname)s with default settings;
    %(classname)s(%(shortvarname)s) : make a copy of %(shortvarname)s;
    %(classname)s(%(shortvarname)s,x=a,...): make a copy of %(shortvarname)s and set property 'x' to a, etc..
    """)

    def _str(self, interactive=False):
        import cStringIO
        sio = cStringIO.StringIO()
        stripProxy(self).printSummaryTree(0, 0, '', out=sio, interactive=interactive)
        returnable = str(sio.getvalue()).rstrip()
        return returnable
    helptext(_str, """Return a printable string representing %(classname)s object as a tree of properties.""")

    def _repr_pretty_(self, p, cycle):
        if cycle:
            p.text('proxy object...')
            return

        p_text = ""
        try:

            if hasattr(self, implRef):
                raw_self = stripProxy(self)
                if hasattr(raw_self, '_repr_pretty_'):
                    raw_self._repr_pretty_(p, cycle)
                elif hasattr(raw_self, '_display'):
                    p_text = raw_self._display()
                else:
                    p_text = self.__str__(True)
            else:
                p_text = self.__str__(True)
        except Exception as err:
            p_text = "Error Representing object: %s\nErr:\n%s" % (type(self), err)

        p.text(p_text)

    helptext(_repr_pretty_, """Return a nice string to be printed in the IPython termial""")

    def _repr(self):
        try:
            has_proxy = hasattr(self, implRef)
            if has_proxy:
                raw_proxy = stripProxy(self)
            else:
                raw_proxy = None
            if has_proxy and hasattr(raw_proxy, '_repr'):
                return raw_proxy._repr()
            else:
                return '<' + repr(stripProxy(self)) + ' PROXY at ' + hex(abs(id(self))) + '>'
        except Exception as err:
            return "Error Representing object: %s\nErr:\n" % (type(self), err)

    helptext(_repr, "Return an short representation of %(classname)s object.")

    def _eq(self, x):
        result = False
        if isType(x, GPIProxyObject) or hasattr(x, implRef):
            result = stripProxy(self).__eq__(stripProxy(x))
        else:
            result = stripProxy(self).__eq__(x)
        return result
    helptext(_eq, "Equality operator (==), compare the %(classname)s properties which are declared as [comparable].")

    def _ne(self, x):
        result = True
        if isType(x, GPIProxyObject) or hasattr(x, implRef):
            result = stripProxy(self).__ne__(stripProxy(x))
        else:
            result = stripProxy(self).__ne__(x)
        return result
    helptext(_ne, "Non-equality operator (!=).")

    def _copy(self, unprepare=None):
        logger.debug('unprepare is %s', unprepare)
        if unprepare is None:
            if prepconfig['unprepare_on_copy'] is True:
                if hasattr(self, 'is_prepared') or hasattr(self, 'application'):
                    unprepare = True

        def _getSharedPath():
            Config_conf = getConfig('Configuration')
            return os.path.join(expandfilename(Config_conf['gangadir']), 'shared', Config_conf['user'])

        if hasattr(self, 'application'):
            if hasattr(self.application, 'is_prepared'):
                from GangaCore.Utility.files import expandfilename
                if self.application.is_prepared not in [None, True]:
                    if hasattr(self.application.is_prepared, 'name'):
                        shared_path = _getSharedPath()
                        if os.path.isdir(os.path.join(shared_path, self.application.is_prepared.name)):
                            from GangaCore.Core.GangaRepository import getRegistry
                            shareref = GPIProxyObjectFactory(getRegistry("prep").getShareRef())
                            logger.debug('increasing counter from proxy.py')
                            shareref.increase(self.application.is_prepared)
                            logger.debug('Found ShareDir directory: %s' % self.application.is_prepared.name)
                elif self.application.is_prepared not in [None, True]:
                    shared_path = _getSharedPath()
                    if not os.path.isdir(os.path.join(shared_path, self.application.is_prepared.name)):
                        logger.error('ShareDir directory not found: %s' % self.application.is_prepared.name)
                        logger.error('Unpreparing Job #%s' % self.id)
                        from GangaCore.Core.GangaRepository import getRegistry
                        shareref = GPIProxyObjectFactory(getRegistry("prep").getShareRef())
                        shareref.increase(self.application.is_prepared)
                        self.unprepare()

        if unprepare is True:
            if hasattr(self, 'is_prepared'):
                from GangaCore.Utility.files import expandfilename
                if self.is_prepared not in [None, True]:
                    if hasattr(self.is_prepared, 'name'):
                        shared_path = _getSharedPath()
                        if not os.path.isdir(os.path.join(shared_path, self.is_prepared.name)):
                            logger.error('ShareDir directory not found: %s' % self.is_prepared.name)
                            logger.error('Unpreparing %s application' % getName(self))
                            self.unprepare()

            c = stripProxy(self).clone()
            if hasattr(c, 'is_prepared') and c._getRegistry() is None:
                from GangaCore.Core.GangaRepository import getRegistry
                shareref = GPIProxyObjectFactory(getRegistry("prep").getShareRef())
                shareref.increase(self.is_prepared)
            stripProxy(c)._auto__init__(unprepare=True)
        else:
            c = stripProxy(self).clone()
            stripProxy(c)._auto__init__()

        return addProxy(c)

    helptext(_copy, "Make an identical copy of self.")

    def _setattr(self, x, v):
        'This is the setter method in the Proxied Objects'
        #logger.debug("_setattr")
        # need to know about the types that require metadata attribute checking
        # this allows derived types to get same behaviour for free.
        raw_self = stripProxy(self)
        p_Ref = raw_self
        if p_Ref is not None:
            if not isclass(p_Ref):
                class_type = type(p_Ref)
            else:
                class_type = p_Ref
        else:
            class_type = p_Ref

        if x == implRef and not isinstance(v, class_type):
            raise AttributeError("Internal implementation object '%s' cannot be reassigned" % implRef)

        elif not raw_self._schema.hasAttribute(x):
            from GangaCore.GPIDev.Lib.Job.MetadataDict import MetadataDict
            if hasattr(raw_self, 'metadata') and isType(raw_self.metadata, MetadataDict):
                if x in raw_self.metadata.data:
                    raise GangaAttributeError("Metadata item '%s' cannot be modified" % x)

            if x != implRef:
                raise GangaAttributeError("Can't assign '%s' as it does NOT appear in the object schema for class '%s'" % (x, getName(self)))

        new_v = stripProxy(runtimeEvalString(self, x, v))
        GPIProxyObject.__setattr__(self, x, new_v)


    helptext(_setattr, """Set a property of %(classname)s with consistency and safety checks.
Setting a [protected] or a unexisting property raises AttributeError.""")

    #    def _getattr(self, name):
#        if name == '_impl': return self._impl
#        if '_attribute_filter__get__' in dir(self._impl):
#            return self._impl._attribute_filter__get__(name)
#        return self.name
#        ## need to know about the types that require metadata attribute checking
#        ## this allows derived types to get same behaviour for free.
#        from GangaCore.GPIDev.Lib.Job.Job import Job
#        from GangaCore.GPIDev.Lib.Tasks.Task import Task
#        from GangaCore.GPIDev.Lib.Tasks.Transform import Transform
#        metadata_objects=[Job]
#        if True in (isType(self,t) for t in metadata_objects):
#            try:
#                return self.metadata[name]
#            except:
#                return object.__getattribute__(self,name)
#        return object.__getattribute__(self,name)

    def _getattribute(self, name):

        #logger.debug("_getattribute: %s" % name)

        if name.startswith('__') or name == implRef:
            return GPIProxyObject.__getattribute__(self, name)
        else:
            implInstance = stripProxy(self)

            if '_attribute_filter__get__' in dir(implInstance) and \
                    not isType(implInstance, ObjectMetaclass) and \
                    implInstance._schema.hasItem(name) and \
                    not implInstance._schema.getItem(name)['hidden']:
                        returnable = addProxy(implInstance._attribute_filter__get__(name))
            else:
                try:
                    returnable = GPIProxyObject.__getattribute__(self, name)
                except AttributeError:
                    raise GangaAttributeError("Object '%s' does not have attribute: '%s'" % (getName(self), name))

        if isType(returnable, GangaObject):
            return addProxy(returnable)
        else:
            return returnable

    # but at the class level _impl is a ganga plugin class
    d = {implRef: pluginclass,
            '__init__': _init,
            '__str__': _str,
            '__repr__': _repr,
            '_repr_pretty_': _repr_pretty_,
            '__eq__': _eq,
            '__ne__': _ne,
            'copy': _copy,
            '__doc__': publicdoc,
            '__setattr__': _setattr,
         #          '__getattr__': _getattr,
            '__getattribute__': _getattribute,
         }

    if not hasattr(pluginclass, '_exportmethods'):
        pluginclass._exportmethods = []

    exported_methods = pluginclass._exportmethods

    # export public methods of this class and also of all the bases
    # this class is scanned last to extract the most up-to-date docstring
    exported_dicts = (b.__dict__ for b in reversed(pluginclass.__mro__))

    for subclass_dict in exported_dicts:
        for k in subclass_dict:
            if getattr(subclass_dict[k], 'exported', False):
                exported_methods.append(k)  # Add all @export'd methods
            if k in exported_methods:

                # Should we expose the object/method directly or a custom '_export_' wrapper?
                internal_name = "_export_" + k
                if internal_name not in subclass_dict:
                    internal_name = k

                try:
                    method = subclass_dict[internal_name]
                except KeyError as err:
                    logger.debug("ObjectMetaClass Error internal_name: %s,\t d: %s" % (internal_name, d))
                    logger.debug("ObjectMetaClass Error: %s" % err)
                    raise

                # If this is a method make sure we wrap the method so it's not exposed to GPI objects
                if not isinstance(method, types.FunctionType):
                    continue

                # Wrap the method so that arguments are stripped before the method runs and a proxied object is returned
                f = ProxyMethodDescriptor(k, internal_name)
                f.__doc__ = method.__doc__
                d[k] = f

    # export visible properties... do not export hidden properties
    for attr, item in pluginclass._schema.allItems():
        if not item['hidden']:
            d[attr] = ProxyDataDescriptor(attr)

    return type(name, (GPIProxyObject,), d)


#
#
# $Log: not supported by cvs2svn $
# Revision 1.2.4.2  2009/07/08 15:39:15  ebke
# removed object_filter__get__
#
# Revision 1.2.4.1  2009/07/08 11:18:21  ebke
# Initial commit of all - mostly small - modifications due to the new GangaRepository.
# No interface visible to the user is changed
#
# Revision 1.2  2008/09/09 14:37:16  moscicki
# bugfix #40220: Ensure that default values satisfy the declared types in the schema
#
# factored out type checking into schema module, fixed a number of wrongly declared schema items in the core
#
# Revision 1.1  2008/07/17 16:40:52  moscicki
# migration of 5.0.2 to HEAD
#
# the doc and release/tools have been taken from HEAD
#
# Revision 1.32.4.15  2008/07/02 13:19:35  moscicki
# TODO comment
#
# Revision 1.32.4.14  2008/06/04 10:34:54  moscicki
# fixed typecheck function naming
#
# Revision 1.32.4.13  2008/06/02 11:42:12  moscicki
# fixed typo...
#
# Revision 1.32.4.12  2008/05/27 18:02:50  kuba
# fixed bug #36836, .ganga.log in afs home directory
#
# Revision 1.32.4.11  2008/02/29 09:16:17  moscicki
# fix from Will Reece
#
# Revision 1.32.4.10  2008/02/28 15:45:30  moscicki
# fixed GangaList typechecking problem
#
# Revision 1.32.4.9  2008/02/06 16:26:30  moscicki
# typelist == None => disable type checking
# extra warnings for incomplete typesystem information
#
# Revision 1.32.4.8  2008/02/06 09:48:44  wreece
# Allows the proxy stripper a NoOp if the object has already been stripped.
#
# Revision 1.32.4.7  2008/02/06 09:28:48  wreece
# First pass at a cleanup of the gangalist stuff. I've made changes so the diffs with the 4.4 series are more transparent. Currently still test failures.
#
# Revision 1.32.4.6  2007/12/18 16:40:39  moscicki
# bugfix
#
# Revision 1.32.4.5  2007/12/18 09:05:03  moscicki
# integrated typesystem from Alvin and made more uniform error reporting
#
# Revision 1.32.4.4  2007/11/20 14:29:49  wreece
# Corrects typo in TFile, and removes extra methods from Proxy. Typo was causing tests to fail.
#
# Revision 1.32.4.3  2007/11/14 13:03:54  wreece
# Changes to make shortcuts work correctly with gangalists. all but one tests should now pass.
#
# Revision 1.32.4.2  2007/11/07 17:02:10  moscicki
# merged against Ganga-4-4-0-dev-branch-kuba-slices with a lot of manual merging
#
# Revision 1.32.4.1  2007/11/07 15:10:02  moscicki
# merged in pretty print and GangaList support from ganga-5-dev-branch-4-4-1-will-print branch
#
# Revision 1.32.8.2  2007/10/30 14:30:23  wreece
# Non-working update. Adds in Kuba's exported methods dodge. It is now possible to define a _export_ version of a method for external use and a undecorated method for internal use.
#
# Revision 1.32.8.1  2007/10/30 12:12:08  wreece
# First version of the new print_summary functionality. Lots of changes, but some known limitations. Will address in next version.
#
# Revision 1.32  2007/07/10 13:08:29  moscicki
# docstring updates (ganga devdays)
#
# Revision 1.31  2007/07/10 07:50:52  moscicki
# fixed #27541
#
# Revision 1.30.8.1  2007/06/18 10:16:34  moscicki
# slices prototype
#
# Revision 1.30  2007/03/07 09:24:34  moscicki
# AGAIN: fixed a problem of assigning iterable plain value (such as string) to a non-strict sequence (so x.files = "abc" was yelding 3 files "a","b","c"
#
# added GangaAttributeError exception
#
# Revision 1.29  2007/03/06 16:38:24  moscicki
# fixed a problem of assigning iterable plain value (such as string) to a non-strict sequence (so x.files = "abc" was yelding 3 files "a","b","c"
#
# Revision 1.28  2007/03/05 12:03:01  moscicki
# explicit switch for strict_sequence (default is True), if the sequence is non-strict then a single value v will be converted to [v] on assignment, for example non-strict File sequence yields obj.x = 'a' <=> obj.x = [File('a')]  <=> obj.x = File('a')
#
# Revision 1.27  2007/02/28 18:20:49  moscicki
# moved GangaException to GangaCore.Core
#
# Revision 1.26  2006/10/23 11:00:42  moscicki
# minor logger fixed (level changed)
#
# Revision 1.25  2006/09/15 14:19:44  moscicki
# Fixed bug #12229 overview: FutureWarning from Python
#
# Revision 1.24  2006/08/11 13:13:05  adim
# Added: GangaException as a markup base class for all exception that need to be printed in a usable way in IPython shell
#
# Revision 1.23  2005/12/02 15:28:46  moscicki
# customizable _repr() method on GPI objects
#
# Revision 1.22  2005/11/14 13:58:27  moscicki
# fixed a getter filter bug
#
# Revision 1.21  2005/11/14 10:30:37  moscicki
# enabled getter attribute filter
#
# Revision 1.20  2005/09/21 09:09:14  moscicki
# "better" repr text for proxies
#
# Revision 1.19  2005/08/26 13:21:47  moscicki
# do not document hidden properties
#
# Revision 1.18  2005/08/26 10:21:31  karl
# KH: Minor correction: linesep -> separator
#
# Revision 1.17  2005/08/26 09:54:55  moscicki
# minor changes
#
# Revision 1.16  2005/08/24 15:41:19  moscicki
# automatically generated help for properties, disabled the SchemaHelper and few other improvements to the help system
#
# Revision 1.15  2005/08/23 17:15:06  moscicki
# *** empty log message ***
#
#
#



################################################################################
# GangaND280 Project.
# Anthony Hillairet
# Created 11/01/2014
################################################################################
"""@package ND280Splitter
This module contains various splitters for ND280 jobs.
"""

import inspect

from GangaCore.GPIDev.Adapters.ISplitter import ISplitter
from GangaCore.GPIDev.Base.Proxy import addProxy, stripProxy
from GangaCore.GPIDev.Schema import *
from GangaCore.Utility.logging import getLogger

logger = getLogger()

# First define the functions that can be used here or in the transforms
def splitCSVFile(csvfile, nbevents):
    subsets = []
    allLines = []
    incsvfile = csvfile
    csvfilebuf = open(incsvfile, 'rb')
    for line in csvfilebuf:
        line = line.rstrip('\r\n')
        row = line.split(",")
        if not len(row) == 3:
            print "Ignoring badly-formatted line:", ",".join(row)
            continue
        allLines.append(line)
    csvfilebuf.close()

    if nbevents < 1:
      raise Exception('Number of nbevents not set properly.')

    subsets = []
    # Less lines than number of events per job wanted => easy
    if len(allLines) < nbevents:
      subsets.append(allLines)
    else:
      nbfulljobs = len(allLines) / nbevents
      for nb in range(nbfulljobs):
        Low = nb * nbevents
        High = (nb+1) * nbevents
        subsets.append(allLines[Low:High])

      if len(allLines) % nbevents:
        # If the number of lines is not divisible by nbevents
        # then the last subjob which has less
        subsets.append(allLines[High:])

    return subsets


def splitNbInputFile(infiles, nbfiles):
    subsets = []
    # For nbfiles 0 or less, just put all the input files in one subset.
    if nbfiles < 1:
        subsets = [infiles]
    # Less files than number of jobs wanted => easy.
    elif len(infiles) < nbfiles:
      for f in infiles:
        subsets.append([f])
    else:
      nbfulljobs = len(infiles) / nbfiles
      for nb in range(nbfulljobs):
        Low = nb*nbfiles
        High = (nb+1)*nbfiles
        subsets.append(infiles[Low:High])

      if len(infiles) % nbfiles:
        # If the number of input files is not divisible by nbfiles
        # then the last subjob which has less
        subsets.append(infiles[High:])

    return subsets


class ND280SplitNbJobs(ISplitter):

    """
    Split an ND280 job into a given number of subjobs.
    The splitting is done such that all subjobs have
    the same number of input except for the last
    subjob which as the same number or less.

    For example, to split a job into N subjobs:

      S = ND280SplitNbJobs()
      S.nbjobs = 10
    or equivalent
      S = ND280SplitNbJobs(nbjobs=10)
      ... ...
      j = Job(splitter=S)
      j.submit()

    """    
    _name = "ND280SplitNbJobs"
    _schema = Schema(Version(1,0), {
        'nbjobs' : SimpleItem(defvalue=-1,doc='The number of subjobs'),
        } )

    def split(self,job):
        
        subjobs = []

        filenames = job.inputdata.get_dataset_filenames()
      
        logger.info('Creating %d subjobs ...',self.nbjobs)

        if self.nbjobs < 1:
          raise Exception('Number of nbjobs not set properly.')

        subsets = []
        # Less files than number of jobs wanted => easy
        if len(filenames) < self.nbjobs:
          for f in filenames:
            subsets.append([f])
        else:
          isPerfectSplit = (len(filenames) % self.nbjobs) == 0
          if isPerfectSplit:
            # If the number of input files is divisible by nbjobs
            # then all subjobs have the same number of input files
            nbfulljobs = self.nbjobs
          else:
            # Otherwise all subjobs have the same number of input files
            # except the last subjob which has less
            nbfulljobs = self.nbjobs - 1

          persub = len(filenames) / nbfulljobs
          for nb in range(nbfulljobs):
            Low = nb*persub
            High = (nb+1)*persub
            subsets.append(filenames[Low:High])

          if not isPerfectSplit:
            subsets.append(filenames[High:])

        for sub in subsets:

            j = addProxy(self.createSubjob(job))

            j.inputdata.set_dataset_filenames(sub)

            subjobs.append(stripProxy(j))

        return subjobs



class ND280SplitNbInputFiles(ISplitter):

    """
    Split job into a number of subjobs such that each
    have the same given number of input files or less.

    For example, to split into subjobs with N input files
    each:

      S = ND280SplitNbInputFiles()
      S.nbfiles = 10
    or equivalent
      S = ND280SplitNbInputFiles(nbfiles=10)
      ... ...
      j = Job(splitter=S)
      j.submit()

    """    
    _name = "ND280SplitNbInputFiles"
    _schema = Schema(Version(1,0), {
        'nbfiles' : SimpleItem(defvalue=-1,doc='The number of input files for each subjobs'),
        } )

    def split(self,job):
        
        subjobs = []

        filenames = job.inputdata.get_dataset_filenames()
      
        if self.nbfiles < 1:
          raise Exception('Number of nbfiles not set properly.')

        subsets = splitNbInputFile(filenames, self.nbfiles)

        logger.info('Creating %d subjobs ...',len(subjobs))

        for sub in subsets:

            j = addProxy(self.createSubjob(job))

            j.inputdata.set_dataset_filenames(sub)

            subjobs.append(stripProxy(j))

        return subjobs




class ND280SplitCSVByNbEvt(ISplitter):

    """
    Split job into a number of subjobs such that each
    subjob has a CSV file with N events.

    For example, to split into subjobs with N events
    each:

      S = ND280SplitCSVByNbEvt()
      S.nbevents = 10
    or equivalent
      S = ND280SplitCSVByNbEvt(nbevents=10)
      ... ...
      j = Job(splitter=S)
      j.submit()

    """    
    _name = "ND280SplitCSVByNbEvt"
    _schema = Schema(Version(1,0), {
        'nbevents' : SimpleItem(defvalue=-1,doc='The number of events for each subjobs'),
        } )

    def split(self,job):
        import os
        
        subjobs = []

        subsets = splitCSVFile(job.application.csvfile, self.nbevents)

        # Less files than number of jobs wanted => easy
        logger.info('Creating %d subjobs ...',len(allLines))

        # Base for the naming of each subjob's CSV file
        tmpname = os.path.basename(incsvfile)
        if len(tmpname.split('.')) > 1:
          patterncsv = '.'.join(tmpname.split('.')[0:-1])+"_sub%d."+ tmpname.split('.')[-1]
        else:
          patterncsv = tmpname+"_sub%d"

        # Base for the naming of each subjob's output file
        tmpname = os.path.basename(job.application.outputfile)
        if len(tmpname.split('.')) > 1:
          patternout = '.'.join(tmpname.split('.')[0:-1])+"_sub%d."+ tmpname.split('.')[-1]
        else:
          patternout = tmpname+"_sub%d"

        for s,sub in enumerate(subsets):
            j = addProxy(self.createSubjob(job))

            j.inputdata = job.inputdata

            subLines = '\n'.join(sub)

            from GangaCore.GPIDev.Lib.File import FileBuffer
            thiscsv = patterncsv % s
            # Save in the main job's inputdir now, then the file will be moved to
            # the inputdir of each subjobs.
            job.getInputWorkspace().writefile(FileBuffer(thiscsv,subLines),executable=0)
            j.application.csvfile = os.path.join(job.inputdir,thiscsv)
            j.application.outputfile = patternout % s

            # Prepare the output filenames which must be unique

            subjobs.append(stripProxy(j))


        return subjobs



class ND280SplitOneInputFile(ISplitter):

    """
    Split job into a number of subjobs such that each
    subjob has one input file.

    For example, to split into subjobs with N input files
    each:

      S = ND280SplitOneInputFile()
      ... ...
      j = Job(splitter=S)
      j.submit()

    """    
    _name = "ND280SplitOneInputFile"
    _schema = Schema(Version(1,0), {
        } )

    def split(self,job):
        
        subjobs = []

        filenames = job.inputdata.get_dataset_filenames()
      
        subsets = []
        # Less files than number of jobs wanted => easy
        logger.info('Creating %d subjobs ...',len(filenames))
        for nb in range(len(filenames)):
            j = addProxy(self.createSubjob(job))

            j.inputdata.set_dataset_filenames([filenames[nb]])

            subjobs.append(stripProxy(j))

        return subjobs



#! /usr/bin/env python

"""\
This script prints out a list of undocumented symbols found in
Python include files, prefixed by their tag kind.

Pass Python's include files to ctags, parse the output into a
dictionary mapping symbol names to tag kinds.

Then, the .tex files from Python docs are read into a giant string.

Finally all symbols not found in the docs are written to standard
output, prefixed with their tag kind.
"""

# Which kind of tags do we need?
TAG_KINDS = "dpst"

# Doc sections to use
DOCSECTIONS = ["api"]# ["api", "ext"]

# Only print symbols starting with this prefix,
# to get all symbols, use an empty string
PREFIXES = ("Py", "PY")

INCLUDEPATTERN = "*.h"

# end of customization section


# Tested with EXUBERANT CTAGS
# see http://ctags.sourceforge.net
#
# ctags fields are separated by tabs.
# The first field is the name, the last field the type:
# d macro definitions (and #undef names)
# e enumerators
# f function definitions
# g enumeration names
# m class, struct, or union members
# n namespaces
# p function prototypes and declarations
# s structure names
# t typedefs
# u union names
# v variable definitions
# x extern and forward variable declarations

import os, glob, re, sys

def findnames(file, prefixes=()):
    names = {}
    for line in file.xreadlines():
        if line[0] == '!':
            continue
        fields = line.split()
        name, tag = fields[0], fields[-1]
        if tag == 'd' and name.endswith('_H'):
            continue
        if prefixes:
            sw = name.startswith
            for prefix in prefixes:
                if sw(prefix):
                    names[name] = tag
        else:
            names[name] = tag
    return names

def print_undoc_symbols(prefix, docdir, incdir):
    docs = []

    for sect in DOCSECTIONS:
        for file in glob.glob(os.path.join(docdir, sect, "*.tex")):
            docs.append(open(file).read())

    docs = "\n".join(docs)

    incfiles = os.path.join(incdir, INCLUDEPATTERN)

    fp = os.popen("ctags -IDL_IMPORT --c-types=%s -f - %s"
                  % (TAG_KINDS, incfiles))
    dict = findnames(fp, prefix)
    names = dict.keys()
    names.sort()
    for name in names:
        if not re.search("%s\\W" % name, docs):
            print dict[name], name

if __name__ == '__main__':
    srcdir = os.path.dirname(sys.argv[0])
    incdir = os.path.normpath(os.path.join(srcdir, "../../Include"))
    docdir = os.path.normpath(os.path.join(srcdir, ".."))

    print_undoc_symbols(PREFIXES, docdir, incdir)



#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#
"""The Ganga backendhandler for the Dirac system."""

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#
import os
import re
import fnmatch
import time
import datetime
import shutil
import tempfile
import math
from collections import defaultdict
from GangaCore.GPIDev.Schema import Schema, Version, SimpleItem, ComponentItem
from GangaCore.GPIDev.Adapters.IBackend import IBackend, group_jobs_by_backend_credential
from GangaCore.GPIDev.Lib.Job.Job import Job
from GangaCore.Core.exceptions import GangaFileError, GangaKeyError, BackendError, IncompleteJobSubmissionError
from GangaDirac.Lib.Backends.DiracUtils import result_ok, get_job_ident, get_parametric_datasets, outputfiles_iterator, outputfiles_foreach, getAccessURLs
from GangaDirac.Lib.Files.DiracFile import DiracFile
from GangaDirac.Lib.Utilities.DiracUtilities import GangaDiracError, execute
from GangaDirac.Lib.Credentials.DiracProxy import DiracProxy
from GangaCore.Utility.ColourText import getColour
from GangaCore.Utility.Config import getConfig
from GangaCore.Utility.logging import getLogger, log_user_exception
from GangaCore.GPIDev.Credentials import require_credential, credential_store, needed_credentials
from GangaCore.GPIDev.Base.Proxy import stripProxy, isType, getName
from GangaCore.Core.GangaThread.WorkerThreads import getQueues
from GangaCore.Core import monitoring_component
from GangaCore.Runtime.GPIexport import exportToGPI
from subprocess import check_output, CalledProcessError
configDirac = getConfig('DIRAC')
default_finaliseOnMaster = configDirac['default_finaliseOnMaster']
default_downloadOutputSandbox = configDirac['default_downloadOutputSandbox']
default_unpackOutputSandbox = configDirac['default_unpackOutputSandbox']
logger = getLogger()
regex = re.compile('[*?\[\]]')

class DiracBase(IBackend):

    """The backend that submits jobs to the Grid via DIRAC.

    The backend for jobs to be submitted to the Grid. Jobs are
    submitted through the DIRAC WMS system and then in turn submitted to the
    Grid. A few examples of usage are given below

    # Create Dirac backend object
    b = Dirac()

    # Create and submit job.
    j = Job(application=app,backend=b)
    j.submit()

    # Run a Root job on the Grid if in LHCb VO

    # Create a Root application object. See Root help text for instructions
    # on how to configure this.
    app = Root()

    # Create and submit job to Dirac using default options
    j = Job(application=app,backend=Dirac())
    j.submit()

    # Using the 'settings' attribute
    j.backend.settings['BannedSites'] = ['LCG.CERN.ch']
    j.resubmit()

    # settings can be set at any time but are only 'respected' during
    # submit and resubmit.

    """

    dirac_monitoring_is_active = True

    _schema = Schema(Version(3, 2), {
        'id': SimpleItem(defvalue=None, protected=1, copyable=0,
                         typelist=['int', 'type(None)'],
                         doc='The id number assigned to the job by the DIRAC WMS. If seeking help'
                         ' on jobs with the Dirac backend, please always report this id '
                         'number in addition to a full description of your problem. The id '
                         'can also be used to further inspect the job at '
                         'https://lhcbweb.pic.es/DIRAC/info/general/diracOverview'),
        'status': SimpleItem(defvalue=None, protected=1, copyable=0,
                             typelist=['str', 'type(None)'],
                             doc='The detailed status as reported by the DIRAC WMS'),
        'actualCE': SimpleItem(defvalue=None, protected=1, copyable=0,
                               typelist=['str', 'type(None)'],
                               doc='The location where the job ran'),
        'normCPUTime': SimpleItem(defvalue=None, protected=1, copyable=0,
                                  typelist=['str', 'type(None)'],
                                  doc='The normalized CPU time reported by the DIRAC WMS'),
        'statusInfo': SimpleItem(defvalue='', protected=1, copyable=0,
                                 typelist=['str', 'type(None)'],
                                 doc='Minor status information from Dirac'),
        'extraInfo': SimpleItem(defvalue='', protected=1, copyable=0,
                                typelist=['str', 'type(None)'],
                                doc='Application status information from Dirac'),
        'diracOpts': SimpleItem(defvalue='',
                                doc='DIRAC API commands to add the job definition script. Only edit '
                                'if you *really* know what you are doing'),
        'settings': SimpleItem(defvalue={'CPUTime': 14 * 86400},
                               doc='Settings for DIRAC job (e.g. CPUTime, BannedSites, etc.)'),
        'credential_requirements': ComponentItem('CredentialRequirement', defvalue=DiracProxy),
        'blockSubmit' : SimpleItem(defvalue=True, 
                               doc='Shall we use the block submission?'),
        'finaliseOnMaster' : SimpleItem(defvalue=default_finaliseOnMaster,
                               doc='Finalise the subjobs all in one go when they are all finished.'),
        'downloadSandbox' : SimpleItem(defvalue=default_downloadOutputSandbox,
                               doc='Do you want to download the output sandbox when the job finalises.'),
        'unpackOutputSandbox' : SimpleItem(defvalue=default_unpackOutputSandbox, hidden=True,
                                           doc='Should the output sandbox be unpacked when downloaded.'),

    })
    _exportmethods = ['getOutputData', 'getOutputSandbox', 'removeOutputData',
                      'getOutputDataLFNs', 'getOutputDataAccessURLs', 'peek', 'reset', 'debug', 'finaliseCompletingJobs']
    _packed_input_sandbox = True
    _category = "backends"
    _name = 'DiracBase'
    _hidden = True

    def _setup_subjob_dataset(self, dataset):
        """
        This method is used for constructing datasets on a per subjob basis when submitting parametric jobs
        Args:
            Dataset (Dataset): This is a GangaDataset object, todo check this isn't a list
        """
        return None

    def _setup_bulk_subjobs(self, dirac_ids, dirac_script):
        """
        This is the old bulk submit method which is used to construct the subjobs for a parametric job
        Args:
            dirac_ids (list): This is a list of the Dirac ids which have been created
            dirac_script (str): Name of the dirac script which contains the job jdl
        """
        f = open(dirac_script, 'r')
        parametric_datasets = get_parametric_datasets(f.read().split('\n'))
        f.close()
        if len(parametric_datasets) != len(dirac_ids):
            raise BackendError('Dirac', 'Missmatch between number of datasets defines in dirac API script and those returned by DIRAC')

        master_job = self.getJobObject()
        master_job.subjobs = []
        for i in range(len(dirac_ids)):
            j = Job()
            j.copyFrom(master_job)
            j.splitter = None
            j.backend.id = dirac_ids[i]
            j.id = i
            j.inputdata = self._setup_subjob_dataset(parametric_datasets[i])
            j.status = 'submitted'
            j.time.timenow('submitted')
            master_job.subjobs.append(j)
        return True

    @require_credential
    def _common_submit(self, dirac_script):
        '''Submit the job via the Dirac server.
        Args:
            dirac_script (str): filename of the JDL which is to be submitted to DIRAC
        '''
        j = self.getJobObject()
        self.id = None
        self.actualCE = None
        self.status = None
        self.extraInfo = None
        self.statusInfo = ''
        j.been_queued = False
        dirac_cmd = """execfile(\'%s\')""" % dirac_script

        try:
            result = execute(dirac_cmd, cred_req=self.credential_requirements)
        except GangaDiracError as err:

            err_msg = 'Error submitting job to Dirac: %s' % str(err)
            logger.error(err_msg)
            logger.error("\n\n===\n%s\n===\n" % dirac_script)
            logger.error("\n\n====\n")
            with open(dirac_script, 'r') as file_in:
                logger.error("%s" % file_in.read())
            logger.error("\n====\n")
            raise BackendError('Dirac', err_msg)

        idlist = result
        if type(idlist) is list:
            return self._setup_bulk_subjobs(idlist, dirac_script)

        self.id = idlist
        return type(self.id) == int

    def _addition_sandbox_content(self, subjobconfig):
        '''any additional files that should be sent to dirac
        Args:
            subjobcofig (unknown): This is the config for this subjob (I think)'''
        return []

    @require_credential
    def _block_submit(self, myscript, lenSubjobs, keep_going = False):
        '''Submit a block of jobs via the Dirac server in one go.
        Args:
            dirac_script (str): filename of the JDL which is to be submitted to DIRAC
        '''
        j = self.getJobObject()
        self.id = None
        self.actualCE = None
        self.status = None
        self.extraInfo = None
        self.statusInfo = ''
        j.been_queued = False
        dirac_cmd = """execfile(\'%s\')""" % myscript
        submitFailures = {}
        try:
            result = execute(dirac_cmd, cred_req=self.credential_requirements, return_raw_dict = True, new_subprocess = True)

        except GangaDiracError as err:
            err_msg = 'Error submitting job to Dirac: %s' % str(err)
            logger.error(err_msg)
            logger.error("\n\n===\n%s\n===\n" % myscript)
            logger.error("\n\n====\n")
            with open(myscript, 'r') as file_in:
                logger.error("%s" % file_in.read())
            logger.error("\n====\n")
            j.updateStatus('failed')
            raise BackendError('Dirac', err_msg)
            return 0

        #Now put the list of Dirac IDs into the subjobs and get them monitored:
        if len(j.subjobs)>0:
            for jobNo in result.keys():
                sjNo = jobNo.split('.')[1]
                #If we get an int we have a DIRAC ID so job submitted
                if isinstance(result[jobNo], int):
                    j.subjobs[int(sjNo)].backend.id = result[jobNo]
                    j.subjobs[int(sjNo)].updateStatus('submitted')
                    j.time.timenow('submitted')
                    stripProxy(j.subjobs[int(sjNo)].info).increment()
                #If we get a string we have an error message. Add to the list of failures
                elif isinstance(result[jobNo], str):
                    j.subjobs[int(sjNo)].updateStatus('failed')
                    submitFailures.update({jobNo : result[jobNo]})
                #If we have neither int or str then something disastrous happened
                else:
                    j.subjobs[int(sjNo)].updateStatus('failed')
                    submitFailures.update({jobNo : 'DIRAC error!'})
        else:
            if isinstance(result[result.keys()[0]], int):
                j.backend.id = result[result.keys()[0]]
                j.updateStatus('submitted')
                j.time.timenow('submitted')
                stripProxy(j.info).increment()
            elif isinstance(result[result.keys()[0]], str):
                j.updateStatus('failed')
                submitFailures.update({result.keys()[0] : result[result.keys()[0]]})
            else:
                j.updateStatus('failed')
                submitFailures.update({result.keys()[0] : 'DIRAC error!'})

        #Check that every subjob got submitted ok
        if len(submitFailures.keys()) > 0:
            for sjNo in submitFailures.keys():
                logger.error('Job submission failed for job %s : %s' % (sjNo, submitFailures[sjNo]))
            raise GangaDiracError("Some subjobs failed to submit! Check their status!")
            return 0

        return 1 

    def master_submit(self, rjobs, subjobconfigs, masterjobconfig, keep_going=False, parallel_submit=False):
        """  Submit the master job and all of its subjobs. To keep things speedy when talking to DIRAC
        we can submit several subjobs in the same process. Therefore for each subjob we collect the code for
       the dirac-script into one large file that we then execute.
        """
        #If you want to go slowly use the regular master_submit:
        if not self.blockSubmit:
            return IBackend.master_submit(self, rjobs, subjobconfigs, masterjobconfig, keep_going, parallel_submit)

        #Otherwise use the block submit. Much of this is copied from IBackend
        logger.debug("SubJobConfigs: %s" % len(subjobconfigs))
        logger.debug("rjobs: %s" % len(rjobs))

        if rjobs and len(subjobconfigs) != len(rjobs):
            raise BackendError("The number of subjob configurations does not match the number of subjobs!")

        incomplete = 0
        incomplete_subjobs = []

        master_input_sandbox = self.master_prepare(masterjobconfig)

        nPerProcess = configDirac['maxSubjobsPerProcess']
        nProcessToUse = math.ceil((len(rjobs)*1.0)/nPerProcess)

        from GangaCore.Core.GangaThread.WorkerThreads import getQueues
        # Must check for credentials here as we cannot handle missing credentials on Queues by design!
        try:
            cred = credential_store[self.credential_requirements]
        except GangaKeyError:
            credential_store.create(self.credential_requirements)

        tmp_dir = tempfile.mkdtemp()
        # Loop over the processes and create the master script for each one.
        for i in range(0,int(nProcessToUse)):
            nSubjobs = 0
            #The Dirac IDs are stored in a dict so create it at the start of the script
            masterScript = 'resultdict = {}\n'
            for sc, sj in zip(subjobconfigs[i*nPerProcess:(i+1)*nPerProcess], rjobs[i*nPerProcess:(i+1)*nPerProcess]):
                #Add in the script for each subjob
                sj.updateStatus('submitting')
                fqid = sj.getFQID('.')
                #Change the output of the job script for our own ends. This is a bit of a hack but it saves having to rewrite every RTHandler
                sjScript = sj.backend._job_script(sc, master_input_sandbox, tmp_dir)
                sjScript = sjScript.replace("output(result)", "if isinstance(result, dict) and 'Value' in result:\n\tresultdict.update({sjNo : result['Value']})\nelse:\n\tresultdict.update({sjNo : result['Message']})")
                if nSubjobs == 0:
                    sjScript = re.sub("(dirac = Dirac.*\(\))",r"\1\nsjNo='%s'\n" % fqid, sjScript)
                if nSubjobs !=0 :
                    sjScript = sjScript.replace("from DIRAC.Core.Base.Script import parseCommandLine\nparseCommandLine()\n", "\n")
                    sjScript = re.sub("from .*DIRAC\.Interfaces\.API.Dirac.* import Dirac.*","",sjScript)
                    sjScript = re.sub("from .*DIRAC\.Interfaces\.API\..*Job import .*Job","",sjScript)
                    sjScript = re.sub("dirac = Dirac.*\(\)","",sjScript)
                    masterScript += "\nsjNo=\'%s\'" % fqid
                masterScript += sjScript
                nSubjobs +=1
            #Return the dict of job numbers and Dirac IDs          
            masterScript += '\noutput(resultdict)\n'
            dirac_script_filename = os.path.join(self.getJobObject().getInputWorkspace().getPath(),'dirac-script-%s.py') % i
            with open(dirac_script_filename, 'w') as f:
                f.write(masterScript)
            upperlimit = (i+1)*nPerProcess
            if upperlimit > len(rjobs) :
                upperlimit = len(rjobs)

            if (upperlimit-1) == 0:
                logger.info("Submitting job")
            else:
                logger.info("Submitting subjobs %s to %s" % (i*nPerProcess, upperlimit-1))

            try:
                #Either do the submission in parallel with threads or sequentially
                if parallel_submit:
                    getQueues()._monitoring_threadpool.add_function(self._block_submit, (dirac_script_filename, nSubjobs, keep_going))
                else:
                    self._block_submit(dirac_script_filename, nSubjobs, keep_going)

                while not self._subjob_status_check(rjobs, nPerProcess, i):
                    time.sleep(1.)
            finally:
                shutil.rmtree(tmp_dir, ignore_errors = True)
                    
            if (upperlimit-1) == 0:
                logger.info("Submitted job")
            else:
                logger.info("Submitted subjobs %s to %s" % (i*nPerProcess, upperlimit-1))

        for i in rjobs:
            if i.status in ["new", "failed"]:
                return 0
    
        return 1

    def _subjob_status_check(self, rjobs, nPerProcess, i):
                has_submitted = True
                for sj in rjobs[i*nPerProcess:(i+1)*nPerProcess]:
                    if sj.status not in ["submitted","failed","completed","running","completing"]:
                        has_submitted = False
                        break
                return has_submitted

    def _job_script(self, subjobconfig, master_input_sandbox, tmp_dir):
        """Get the script to submit a single DIRAC job
        Args:
            subjobconfig (unknown):
            master_input_sandbox (list): file names which are in the master sandbox of the master sandbox (if any)
            tmp_dir: (str) This is the temp directory where files will be placed when needed
        """

        j = self.getJobObject()

        sboxname = j.createPackedInputSandbox(subjobconfig.getSandboxFiles())

        input_sandbox = master_input_sandbox[:]
        input_sandbox += sboxname

        input_sandbox += self._addition_sandbox_content(subjobconfig)

        lfns = []

        ## Add LFN to the inputfiles section of the file
        if j.master:
            for this_file in j.master.inputfiles:
                if isType(this_file, DiracFile):
                    lfns.append('LFN:'+str(this_file.lfn))
        for this_file in j.inputfiles:
            if isType(this_file, DiracFile):
                lfns.append('LFN:'+str(this_file.lfn))

        # Make sure we only add unique LFN
        input_sandbox += set(lfns)

        # Remove duplicates incase the LFN have also been added by any prior step
        input_sandbox = list(set(input_sandbox))

        logger.debug("dirac_script: %s" % str(subjobconfig.getExeString()))
        logger.debug("sandbox_cont:\n%s" % str(input_sandbox))


        # This is a workaroud for the fact DIRAC doesn't like whitespace in sandbox filenames
        ### START_WORKAROUND

        # Loop through all files and if the filename contains a ' ' copy it to a location which doesn't contain one.
        # This does have the limitation that all file basenames must not contain a ' ' character.
        # However we don't make any in Ganga as of 20/09/16
        sandbox_str = '['
        for file_ in input_sandbox:
            if ' ' in str(file_):
                new_name = os.path.join(tmp_dir, os.path.basename(file_))
                shutil.copy(file_, new_name)
                file_ = new_name
            sandbox_str += '\'' + str(file_) + '\', '
        sandbox_str += ']'
        logger.debug("sandbox_str: %s" % sandbox_str)
        ### FINISH_WORKAROUND

        dirac_script = subjobconfig.getExeString().replace('##INPUT_SANDBOX##', sandbox_str)

        return dirac_script
        
    def submit(self, subjobconfig, master_input_sandbox):
        """Submit a DIRAC job
        Args:
            subjobconfig (unknown):
            master_input_sandbox (list): file names which are in the master sandbox of the master sandbox (if any)
        """

        j = self.getJobObject()

        tmp_dir = tempfile.mkdtemp()
        dirac_script = self._job_script(subjobconfig, master_input_sandbox, tmp_dir)

        dirac_script_filename = os.path.join(j.getInputWorkspace().getPath(), 'dirac-script.py')
        with open(dirac_script_filename, 'w') as f:
            f.write(dirac_script)

        try:
            return self._common_submit(dirac_script_filename)
        finally:
            # CLEANUP after workaround
            shutil.rmtree(tmp_dir, ignore_errors = True)

    def master_auto_resubmit(self, rjobs):
        '''Duplicate of the IBackend.master_resubmit but hooked into auto resubmission
        such that the monitoring server is used rather than the user server
        Args:
            rjobs (list): This is a list of jobs which are to be auto-resubmitted'''

        incomplete = 0

        def handleError(x):
            if incomplete:
                raise x
            else:
                return 0
        try:
            for sj in rjobs:
                fqid = sj.getFQID('.')
                logger.info("resubmitting job %s to %s backend", fqid, getName(sj.backend))
                try:
                    b = sj.backend
                    sj.updateStatus('submitting')
                    if self.blockSubmit:
                        result = b._blockResubmit()
                    else:
                        result = b._resubmit()
                    if result:
                        sj.updateStatus('submitted')
                        # sj._commit() # PENDING: TEMPORARY DISABLED
                        incomplete = 1
                    else:
                        return handleError(IncompleteJobSubmissionError(fqid, 'resubmission failed'))
                except Exception as x:
                    log_user_exception(logger, debug=isType(x, GangaDiracError))
                    return handleError(IncompleteJobSubmissionError(fqid, str(x)))
        finally:
            master = self.getJobObject().master
            if master:
                master.updateMasterJobStatus()
        return 1

    def resubmit(self):
        """Resubmit a DIRAC job"""
        if self.blockSubmit:
            return self._blockResubmit()
        else:
            return self._resubmit()

    def _resubmit(self):
        """Resubmit a DIRAC job"""
        j = self.getJobObject()
        parametric = False
        script_path = os.path.join(j.getInputWorkspace().getPath(), 'dirac-script.py')
        # Check old script
        if j.master is None and not os.path.exists(script_path):
            raise BackendError('Dirac', 'No "dirac-script.py" found in j.inputdir')

        if j.master is not None and not os.path.exists(script_path):
            script_path = os.path.join(
                j.master.getInputWorkspace().getPath(), 'dirac-script.py')
            if not os.path.exists(script_path):
                raise BackendError('Dirac', 'No "dirac-script.py" found in j.inputdir or j.master.inputdir')
            parametric = True

        # Read old script
        f = open(script_path, 'r')
        script = f.read()
        f.close()

        # Create new script - ##note instead of using get_parametric_dataset
        # could just use j.inputdata.
        if parametric is True:
            parametric_datasets = get_parametric_datasets(script.split('\n'))
            if j.master:
                if len(parametric_datasets) != len(j.master.subjobs):
                    raise BackendError('Dirac', 'number of parametric datasets defined in API script doesn\'t match number of master.subjobs')
            if j.inputdata and len(j.inputdata) > 0:
                _input_files = [f for f in j.inputdata if not isType(f, DiracFile)]
            else:
                _input_files = []
            if set(parametric_datasets[j.id]).symmetric_difference(set([f.namePattern for f in _input_files])):
                raise BackendError(
                    'Dirac', 'Mismatch between dirac-script and job attributes.')
            script = script.replace('.setParametricInputData(%s)' % str(parametric_datasets),
                                    '.setInputData(%s)' % str(parametric_datasets[j.id]))
            script = script.replace('%n', str(j.id))  # name

        start_user_settings = '# <-- user settings\n'
        new_script = script[
            :script.find(start_user_settings) + len(start_user_settings)]

        job_ident = get_job_ident(script.split('\n'))
        for key, value in self.settings.iteritems():
            if str(key).startswith('set'):
                _key = key[3:]
            else:
                _key = key
            if type(value) is str:
                template = '%s.set%s("%s")\n'
            else:
                template = '%s.set%s(%s)\n'
            new_script += template % (job_ident, str(_key), str(value))
        new_script += script[script.find('# user settings -->'):]

        # Save new script
        new_script_filename = os.path.join(j.getInputWorkspace().getPath(), 'dirac-script.py')
        f = open(new_script_filename, 'w')
        f.write(new_script)
        f.flush()
        f.close()
        return self._common_submit(new_script_filename)

    def _blockResubmit(self):
        """Resubmit a DIRAC job that was submitted with bulk submission. This requires writing a new dirac-script for the individual job."""
        j = self.getJobObject()
        parametric = False

        if j.master is None:
            scriptDir = j.getInputWorkspace().getPath()
        else:
            scriptDir = j.master.getInputWorkspace().getPath()

        diracScriptFiles = []
        for fileName in os.listdir(scriptDir):
            if fnmatch.fnmatch(fileName, 'dirac-script-*.py'):
                diracScriptFiles.append(fileName)

        #Did we find any of the new style dirac scripts? If not try the old way as the job may have been submitted with an old ganga version.
        if diracScriptFiles == []:
            return self._resubmit()

        new_script_filename = ''

        for diracScript in diracScriptFiles:
            script_path = os.path.join(scriptDir, diracScript)
            # Check old script
            if not os.path.exists(script_path):
                raise BackendError('Dirac', 'No "dirac-script.py" found in j.inputdir')

            # Read old script
            with open(script_path, 'r') as f:
                script = f.read()
            # Is the subjob we want in there?
            if not ("sjNo='%s'" % j.fqid) in script:
                continue

            #First pick out the imports etc at the start
            newScript =  re.compile(r'%s.*?%s' % ('resultdict = {}',"dirac = Dirac.*?\(\)\n"),re.S).search(script).group(0)
            newScript += '\n'
            #Now pick out the job part
            start = "sjNo='%s'" % j.fqid
            #Check if the original script included the check for the dirac output
            if "result[\'Message\']" in script:
                newScript += re.compile(r'%s.*?%s' % (start,"resultdict.update\({sjNo : result\['Message'\]}\)"),re.S).search(script).group(0)
            else:
                newScript += re.compile(r'%s.*?%s' % (start,"resultdict.update\({sjNo : result\['Value'\]}\)"),re.S).search(script).group(0)
            newScript += '\noutput(resultdict)'
            
            # Modify the new script with the user settings

            start_user_settings = '# <-- user settings\n'
            new_script = newScript[
                :newScript.find(start_user_settings) + len(start_user_settings)]

            job_ident = get_job_ident(newScript.split('\n'))
            for key, value in self.settings.iteritems():
                if str(key).startswith('set'):
                    _key = key[3:]
                else:
                    _key = key
                if type(value) is str:
                    template = '%s.set%s("%s")\n'
                else:
                    template = '%s.set%s(%s)\n'
                new_script += template % (job_ident, str(_key), str(value))
            new_script += newScript[newScript.find('# user settings -->'):]

            # Save new script
            new_script_filename = os.path.join(j.getInputWorkspace().getPath(), 'dirac-script.py')
            with open(new_script_filename, 'w') as f:
                f.write(new_script)
            # Break the loop now we have written the new script
            break

        if new_script_filename == '':
            raise BackendError('Dirac', 'Script for job number %s not found. Resubmission failed.' % j.fqid)

        return self._block_submit(new_script_filename, 1)
 
    def reset(self, doSubjobs=False):
        """Resets the state of a job back to 'submitted' so that the
        monitoring will run on it again.
        Args:
            doSubjobs (bool): Should we rest the subjobs associated with this job or not"""
        j = self.getJobObject()

        disallowed = ['submitting', 'killed']
        if j.status in disallowed:
            logger.warning("Can not reset a job in status '%s'." % j.status)
        else:
            j.getOutputWorkspace().remove(preserve_top=True)
            self.extraInfo = None
            self.statusInfo = ''
            self.status = None
            self.actualCE = None
            j.been_queued = False
            j.updateStatus('submitted')
            if j.subjobs and not doSubjobs:
                logger.info('This job has subjobs, if you would like the backends '
                            'of all the subjobs that are in status=\'completing\' or '
                            'status=\'failed\' also reset then recall reset with the '
                            'arg \'True\' i.e. job(3).backend.reset(True)')
            elif j.subjobs and doSubjobs:
                logger.info('resetting the backends of \'completing\' and \'failed\' subjobs.')
                for sj in j.subjobs:
                    if sj.status == 'completing' or sj.status == 'failed':
                        sj.backend.reset()
            if j.master:
                j.master.updateMasterJobStatus()

    @require_credential
    def kill(self):
        """ Kill a Dirac jobs"""
        if not self.id:
            return None
        dirac_cmd = 'kill(%d)' % self.id
        try:
            result = execute(dirac_cmd, cred_req=self.credential_requirements)
        except GangaDiracError as err:
            raise BackendError('Dirac', 'Could not kill job: %s' % err)
        return True

    @require_credential
    def master_kill(self):
        """ A method for killing a job and all of its subjobs. Does it en masse
        for maximum speed.
        """
        j = self.getJobObject()
        if len(j.subjobs)==0:
            return self.kill()
        else:
            kill_list = []
            for sj in j.subjobs:
                if sj.status not in ['completed', 'failed']:
                    kill_list.append(sj.backend.id)
            dirac_cmd = 'kill(%s)' % kill_list
            try:
                result = execute(dirac_cmd, cred_req=self.credential_requirements)
            except GangaDiracError as err:
                raise BackendError('Dirac', 'Dirac failed to kill job %d.' % j.id)
        return True

    @require_credential
    def peek(self, filename=None, command=None):
        """Peek at the output of a job (Note: filename/command are ignored).
        Args:
            filename (str): Ignored but is filename of a file in the sandbox
            command (str): Ignored but is a command which could be executed"""
        dirac_cmd = 'peek(%d)' % self.id
        try:
            result = execute(dirac_cmd, cred_req=self.credential_requirements)
            logger.info(result)
        except GangaDiracError:
            logger.error("No peeking available for Dirac job '%i'.", self.id)

    @require_credential
    def getOutputSandbox(self, outputDir=None, unpack=True):
        """Get the outputsandbox for the job object controlling this backend
        Args:
            outputDir (str): This string represents the output dir where the sandbox is to be placed
        """
        j = self.getJobObject()
        if outputDir is None:
            outputDir = j.getOutputWorkspace().getPath()
        dirac_cmd = "getOutputSandbox(%d,'%s', %s)"  % (self.id, outputDir, unpack)
        try:
            result = execute(dirac_cmd, cred_req=self.credential_requirements)
        except GangaDiracError as err:
            msg = 'Problem retrieving output: %s' % str(err)
            logger.warning(msg)
            return False

        return True

    def removeOutputData(self):
        """
        Remove all the LFNs associated with this job.
        """
        # Note when the API can accept a list for removeFile I will change
        # this.
        j = self.getJobObject()
        logger.info("Removing all DiracFile output for job %s" % j.id)
        lfnsToRemove = []
        if j.subjobs:
            for sj in j.subjobs:
                outputfiles_foreach(sj, DiracFile, lambda x: lfnsToRemove.append(x.lfn))
        else:
            outputfiles_foreach(j, DiracFile, lambda x: lfnsToRemove.append(x.lfn))
        dirac_cmd = "removeFile(%s)" % lfnsToRemove
        try:
            result = execute(dirac_cmd, cred_req=self.credential_requirements)
        except GangaDiracError as err:
            msg = 'Problem removing files: %s' % str(err)
            logger.warning(msg)
            return False
        def clearFileInfo(f):
            f.lfn = ""
            f.locations = []
            f.guid = ''
        if j.subjobs:
            for sj in j.subjobs:
                outputfiles_foreach(sj, DiracFile, lambda x: clearFileInfo(x))
        else:
            outputfiles_foreach(j, DiracFile, lambda x: clearFileInfo(x))

    def getOutputData(self, outputDir=None, names=None, force=False):
        """Retrieve data stored on SE to dir (default=job output workspace).
        If names=None, then all outputdata is downloaded otherwise names should
        be a list of files to download. If force=True then data will be redownloaded
        even if the file already exists.

        Note that if called on a master job then all subjobs outputwill be downloaded.
        If dir is None then the subjobs output goes into their individual
        outputworkspaces as expected. If however one specifies a dir then this is
        treated as a top dir and a subdir for each job will be created below it. This
        will avoid overwriting files with the same name from each subjob.
        Args:
            outputDir (str): This string represents the output dir where the sandbox is to be placed
            names (list): list of names which match namePatterns in the outputfiles
            force (bool): Force the download out data potentially overwriting existing objects
        """
        j = self.getJobObject()
        if outputDir is not None and not os.path.isdir(outputDir):
            raise GangaDiracError("Designated outupt path '%s' must exist and be a directory" % outputDir)

        def download(dirac_file, job, is_subjob=False):
            dirac_file.localDir = job.getOutputWorkspace().getPath()
            if outputDir is not None:
                output_dir = outputDir
                if is_subjob:
                    output_dir = os.path.join(outputDir, job.fqid)
                    if not os.path.isdir(output_dir):
                        os.mkdir(output_dir)
                dirac_file.localDir = output_dir
            if os.path.exists(os.path.join(dirac_file.localDir, os.path.basename(dirac_file.lfn))) and not force:
                return
            try:
                dirac_file.get()
                return dirac_file.lfn
            # should really make the get method throw if doesn't suceed. todo
            except (GangaDiracError, GangaFileError) as e:
                logger.warning(e)

        suceeded = []
        if j.subjobs:
            for sj in j.subjobs:
                suceeded.extend([download(f, sj, True) for f in outputfiles_iterator(sj, DiracFile) if f.lfn != '' and (names is None or f.namePattern in names)])
        else:
            suceeded.extend([download(f, j, False) for f in outputfiles_iterator(j, DiracFile) if f.lfn != '' and (names is None or f.namePattern in names)])

        return filter(lambda x: x is not None, suceeded)

    def getOutputDataLFNs(self):
        """Retrieve the list of LFNs assigned to outputdata"""
        j = self.getJobObject()
        lfns = []

        if j.subjobs:
            for sj in j.subjobs:
                lfns.extend([f.lfn for f in outputfiles_iterator(sj, DiracFile) if f.lfn != ''])
        else:
            lfns.extend([f.lfn for f in outputfiles_iterator(j, DiracFile) if f.lfn != ''])
        return lfns

    def getOutputDataAccessURLs(self):
        """Retrieve the list of accessURLs assigned to outputdata for a job"""
        return getAccessURLs(self.getOutputDataLFNs())

    @require_credential
    def debug(self):
        '''Obtains some (possibly) useful DIRAC debug info. '''
        # check services
        cmd = 'getServicePorts()'
        try:
            result = execute(cmd, cred_req=self.credential_requirements)
        except GangaDiracError as err:
            logger.warning('Could not obtain services: %s' % str(err))
            return
        services = result
        for category in services:
            system, service = category.split('/')
            cmd = "ping('%s','%s')" % (system, service)
            try:
                result = execute(cmd, cred_req=self.credential_requirements)
                msg = 'OK.'
            except GangaDiracError as err:
                msg = '%s' % err
            logger.info('%s: %s' % (category, msg))

        # get pilot info for this job
        if not isinstance(self.id, int):
            return
        j = self.getJobObject()
        cwd = os.getcwd()
        debug_dir = j.getDebugWorkspace().getPath()
        cmd = "getJobPilotOutput(%d,'%s')" % (self.id, debug_dir)
        try:
            result = execute(cmd, cred_req=self.credential_requirements)
            logger.info('Pilot Info: %s/pilot_%d/std.out.' % (debug_dir, self.id))
        except GangaDiracError as err:
            logger.error("%s" % err)

    def finaliseCompletingJobs(self, downloadSandbox=True):
        """
         A function to finalise all the subjobs in the completing state, so they are ready, before all the subjobs complete.
        """
        j = self.getJobObject()
        if j.master:
            j = j.master
        if not j.subjobs:
            logger.warning("There are no subjobs - this will finalise in its own time.")
            return
        jobList = []
        for sj in j.subjobs:
            if sj.status == 'completing':
                jobList.append(sj)
        if len(jobList) == 0:
            logger.warning("No subjobs are ready to be finalised yet. Be more patient.")
            return
        else:
            DiracBase.finalise_jobs(jobList, downloadSandbox)

    @staticmethod
    def _bulk_updateStateTime(jobStateDict, bulk_time_lookup={} ):
        """ This performs the same as the _getStateTime method but loops over a list of job ids within the DIRAC namespace (much faster)
        Args:
            jobStateDict (dict): This is a dict of {job.backend.id : job_status, } elements
            bulk_time_lookup (dict): Dict of result of multiple calls to getBulkStateTime, performed in advance
        """
        for this_state, these_jobs in jobStateDict.iteritems():
            if bulk_time_lookup == {} or this_state not in bulk_time_lookup:
                bulk_result = execute("getBulkStateTime(%s,\'%s\')" % (repr([j.backend.id for j in these_jobs]), this_state), cred_req=these_jobs[0].backend.credential_requirements)  # TODO split jobs by cred_req
            else:
                bulk_result = bulk_time_lookup[this_state]
            for this_job in jobStateDict[this_state]:
                backend_id = this_job.backend.id
                if backend_id in bulk_result and bulk_result[backend_id]:
                    DiracBase._getStateTime(this_job, this_state, {this_state : bulk_result[backend_id]})
                else:
                    DiracBase._getStateTime(this_job, this_state)

    @staticmethod
    def _getStateTime(job, status, getStateTimeResult={}):
        """Returns the timestamps for 'running' or 'completed' by extracting
        their equivalent timestamps from the loggingInfo.
        Args:
            job (Job): This is the job object we want to update
            status (str): This is the Ganga status we're updating (running, completed... etc)
            getStateTimeResult (dict): This is the optional result of executing the approriate getStateTime
                                        against this job.backend.id, if not provided the command is called internally
        """
        # Now private to stop server cross-talk from user thread. Since updateStatus calles
        # this method whether called itself by the user thread or monitoring thread.
        # Now don't use hook but define our own private version
        # used in monitoring loop... messy but works.
        if job.status != status:
            b_list = ['running', 'completing', 'completed', 'failed']
            backend_final = ['failed', 'completed']
            # backend stamps
            if not job.subjobs and status in b_list:
                for childstatus in b_list:
                    if job.backend.id:
                        logger.debug("Accessing getStateTime() in diracAPI")
                        if childstatus in backend_final:
                            if childstatus in getStateTimeResult:
                                be_statetime = getStateTimeResult[childstatus]
                            else:
                                be_statetime = execute("getStateTime(%d,\'%s\')" % (job.backend.id, childstatus), cred_req=job.backend.credential_requirements)
                            job.time.timestamps["backend_final"] = be_statetime
                            logger.debug("Wrote 'backend_final' to timestamps.")
                            break
                        else:
                            time_str = "backend_" + childstatus
                            if time_str not in job.time.timestamps:
                                if childstatus in getStateTimeResult:
                                    be_statetime = getStateTimeResult[childstatus]
                                else:
                                    be_statetime = execute("getStateTime(%d,\'%s\')" % (job.backend.id, childstatus), cred_req=job.backend.credential_requirements)
                                job.time.timestamps["backend_" + childstatus] = be_statetime
                            logger.debug("Wrote 'backend_%s' to timestamps.", childstatus)
                    if childstatus == status:
                        break
            logger.debug("_getStateTime(job with id: %d, '%s') called.", job.id, job.status)
        else:
            logger.debug("Status changed from '%s' to '%s'. No new timestamp was written", job.status, status)

    def timedetails(self):
        """Prints contents of the loggingInfo from the Dirac API."""
        if not self.id:
            return None
        logger.debug("Accessing timedetails() in diracAPI")
        dirac_cmd = 'timedetails(%d)' % self.id
        return execute(dirac_cmd, cred_req=self.credential_requirements)

    @staticmethod
    def job_finalisation_cleanup(job, updated_dirac_status):
        """
        Method for reverting a job back to a clean state upon a failure in the job progression
        Args:
            job (Job) This is the job to change the status
            updated_dirac_status (str): Ganga status which is to be used somewhere
        """
        #   Revert job back to running state if we exit uncleanly
        if job.status == "completing":
            job.updateStatus('running')
            if job.master:
                job.master.updateMasterJobStatus()
        # FIXME should I add something here to cleanup on sandboxes pulled from
        # malformed job output?

    @staticmethod
    def _internal_job_finalisation(job, updated_dirac_status):
        """
        This method performs the main job finalisation
        Args:
            job (Job): Thi is the job we want to finalise
            updated_dirac_status (str): String representing the Ganga finalisation state of the job failed/completed
        """
        if job.backend.finaliseOnMaster and job.master and updated_dirac_status == 'completed':
            job.updateStatus('completing')
            allComplete = True
            jobsToFinalise = []
            for sj in job.master.subjobs:
                if sj.status not in ['completing', 'failed', 'killed', 'removed', 'completed']:
                    allComplete = False
                    break
            if allComplete:
                DiracBase.finalise_jobs(job.master.subjobs, job.master.backend.downloadSandbox)
            return

        if updated_dirac_status == 'completed':
            start = time.time()
            # firstly update job to completing
            DiracBase._getStateTime(job, 'completing')
            if job.status in ['removed', 'killed']:
                return
            elif (job.master and job.master.status in ['removed', 'killed']):
                return  # user changed it under us

            job.updateStatus('completing')
            if job.master:
                job.master.updateMasterJobStatus()

            output_path = job.getOutputWorkspace().getPath()

            logger.debug('Contacting DIRAC for job: %s' % job.fqid)
            # Contact dirac which knows about the job
            job.backend.normCPUTime, getSandboxResult, file_info_dict, completeTimeResult = execute("finished_job(%d, '%s', %s, downloadSandbox=%s)" % (job.backend.id, output_path, job.backend.unpackOutputSandbox, job.backend.downloadSandbox), cred_req=job.backend.credential_requirements)

            now = time.time()
            logger.debug('%0.2fs taken to download output from DIRAC for Job %s' % ((now - start), job.fqid))

            #logger.info('Job ' + job.fqid + ' OutputDataInfo: ' + str(file_info_dict))
            #logger.info('Job ' + job.fqid + ' OutputSandbox: ' + str(getSandboxResult))
            #logger.info('Job ' + job.fqid + ' normCPUTime: ' + str(job.backend.normCPUTime))

            # Set DiracFile metadata
            if hasattr(job.outputfiles, 'get'):
                wildcards = [f.namePattern for f in job.outputfiles.get(DiracFile) if regex.search(f.namePattern) is not None]
            else:
                wildcards = []

            lfn_store = os.path.join(output_path, getConfig('Output')['PostProcessLocationsFileName'])

            # Make the file on disk with a nullop...
            if not os.path.isfile(lfn_store):
                with open(lfn_store, 'w'):
                    pass

            if hasattr(job.outputfiles, 'get') and job.outputfiles.get(DiracFile):

                # Now we can iterate over the contents of the file without touching it
                with open(lfn_store, 'ab') as postprocesslocationsfile:
                    if not hasattr(file_info_dict, 'keys'):
                        logger.error("Error understanding OutputDataInfo: %s" % str(file_info_dict))
                        raise GangaDiracError("Error understanding OutputDataInfo: %s" % str(file_info_dict))

                    ## Caution is not clear atm whether this 'Value' is an LHCbism or bug
                    list_of_files = file_info_dict.get('Value', file_info_dict.keys())

                    for file_name in list_of_files:
                        file_name = os.path.basename(file_name)
                        info = file_info_dict.get(file_name)
                        #logger.debug("file_name: %s,\tinfo: %s" % (str(file_name), str(info)))

                        if not hasattr(info, 'get'):
                            logger.error("Error getting OutputDataInfo for: %s" % str(job.getFQID('.')))
                            logger.error("Please check the Dirac Job still exists or attempt a job.backend.reset() to try again!")
                            logger.error("Err: %s" % str(info))
                            logger.error("file_info_dict: %s" % str(file_info_dict))
                            raise GangaDiracError("Error getting OutputDataInfo")

                        valid_wildcards = [wc for wc in wildcards if fnmatch.fnmatch(file_name, wc)]
                        if not valid_wildcards:
                            valid_wildcards.append('')

                        for wc in valid_wildcards:
                            #logger.debug("wildcard: %s" % str(wc))

                            DiracFileData = 'DiracFile:::%s&&%s->%s:::%s:::%s\n' % (wc,
                                                                                    file_name,
                                                                                    info.get('LFN', 'Error Getting LFN!'),
                                                                                    str(info.get('LOCATIONS', ['NotAvailable'])),
                                                                                    info.get('GUID', 'NotAvailable')
                                                                                    )
                            #logger.debug("DiracFileData: %s" % str(DiracFileData))
                            postprocesslocationsfile.write(DiracFileData)
                            postprocesslocationsfile.flush()

                logger.debug("Written: %s" % open(lfn_store, 'r').readlines())

            # check outputsandbox downloaded correctly
            if job.backend.downloadSandbox and not result_ok(getSandboxResult):
                logger.warning('Problem retrieving outputsandbox: %s' % str(getSandboxResult))
                DiracBase._getStateTime(job, 'failed')
                if job.status in ['removed', 'killed']:
                    return
                elif (job.master and job.master.status in ['removed', 'killed']):
                    return  # user changed it under us
                job.updateStatus('failed')
                if job.master:
                    job.master.updateMasterJobStatus()
                raise BackendError('Dirac', 'Problem retrieving outputsandbox: %s' % str(getSandboxResult))
            #If the sandbox dict includes a Succesful key then the sandbox has been download from grid storage, likely due to being oversized. Untar it and issue a warning.
            elif job.backend.downloadSandbox and isinstance(getSandboxResult['Value'], dict) and getSandboxResult['Value'].get('Successful', False):
                    try:
                        sandbox_name = getSandboxResult['Value']['Successful'].values()[0]
                        check_output(['tar', '-xvf', sandbox_name, '-C', output_path])
                        check_output(['rm', sandbox_name])
                        logger.warning('Output sandbox for job %s downloaded from grid storage due to being oversized.' % job.fqid)
                    except CalledProcessError:
                        logger.error('Failed to unpack output sandbox for job %s' % job.fqid)
            # finally update job to completed
            DiracBase._getStateTime(job, 'completed', completeTimeResult)
            if job.status in ['removed', 'killed']:
                return
            elif (job.master and job.master.status in ['removed', 'killed']):
                return  # user changed it under us
            job.updateStatus('completed')
            if job.master:
                job.master.updateMasterJobStatus()
            now = time.time()
            logger.debug('Job ' + job.fqid + ' Time for complete update : ' + str(now - start))

        elif updated_dirac_status == 'failed':
            # firstly update status to failed
            DiracBase._getStateTime(job, 'failed')
            if job.status in ['removed', 'killed']:
                return
            if (job.master and job.master.status in ['removed', 'killed']):
                return  # user changed it under us
            job.updateStatus('failed')
            if job.master:
                job.master.updateMasterJobStatus()

            # if requested try downloading outputsandbox anyway
            if configDirac['failed_sandbox_download']:
                execute("getOutputSandbox(%d,'%s', %s)" % (job.backend.id, job.getOutputWorkspace().getPath(), job.backend.unpackOutputSandbox), cred_req=job.backend.credential_requirements)
        else:
            logger.error("Job #%s Unexpected dirac status '%s' encountered" % (job.getFQID('.'), updated_dirac_status))

    @staticmethod
    def job_finalisation(job, updated_dirac_status):
        """
        Attempt to finalise the job given and auto-retry 5 times on error
        Args:
            job (Job): Job object to finalise
            updated_dirac_status (str): The Ganga status to update the job to, i.e. failed/completed
        """
        count = 1
        limit = 5
        sleep_length = 2.5

        while count != limit:

            try:
                count += 1
                # Check status is sane before we start
                if job.status != "running" and (not job.status in ['completed', 'killed', 'removed']):
                    job.updateStatus('submitted')
                    job.updateStatus('running')
                if job.status in ['completed', 'killed', 'removed']:
                    break
                DiracBase._internal_job_finalisation(job, updated_dirac_status)
                break
            except Exception as err:

                logger.warning("An error occured finalising job: %s" % job.getFQID('.'))
                logger.warning("Attemting again (%s of %s) after %s-sec delay" % (str(count), str(limit), str(sleep_length)))
                if count == limit:
                    logger.error("Unable to finalise job after %s retries due to error:\n%s" % (job.getFQID('.'), str(err)))
                    job.force_status('failed')
                    raise

            time.sleep(sleep_length)

        job.been_queued = False

    @staticmethod
    def finalise_jobs(allJobs, downloadSandbox = True):
        """
        Finalise the jobs given. This downloads the output sandboxes, gets the final Dirac stati, completion times etc.
        Everything is done in one DIRAC process for maximum speed. This is also done in parallel for maximum speed.
        """
        theseJobs = []

        for sj in allJobs:
            if stripProxy(sj).status in ['completing', 'failed', 'killed', 'removed']:
                theseJobs.append(sj)
            else:
                logger.warning("Job %s cannot be finalised" % stripProxy(sj).getFQID())
                
        if len(theseJobs) == 0:
            logger.warning("No jobs from the list are ready to be finalised yet. Be more patient.")
            return

        #First grab all the info from Dirac
        inputDict = {}
        #I have to reduce the no. of subjobs per process to prevent DIRAC timeouts
        nPerProcess = int(math.floor(configDirac['maxSubjobsFinalisationPerProcess']))
        nProcessToUse = math.ceil((len(theseJobs)*1.0)/nPerProcess)

        jobs = [stripProxy(j) for j in theseJobs]

        for i in range(0,int(nProcessToUse)):
            jobSlice = jobs[i*nPerProcess:(i+1)*nPerProcess]      
            getQueues()._monitoring_threadpool.add_function(DiracBase.finalise_jobs_thread_func, (jobSlice, downloadSandbox))

    @staticmethod
    def finalise_jobs_thread_func(jobSlice, downloadSandbox = True):
        """
        Finalise the jobs given. This downloads the output sandboxes, gets the final Dirac statuses, completion times etc.
        Everything is done in one DIRAC process for maximum speed. This is also done in parallel for maximum speed.
        """
        inputDict = {}

        for sj in jobSlice:
            inputDict[sj.backend.id] = sj.getOutputWorkspace().getPath()
        statusmapping = configDirac['statusmapping']
        returnDict, statusList = execute("finaliseJobs(%s, %s, %s)" % (inputDict, repr(statusmapping), downloadSandbox), cred_req=jobSlice[0].backend.credential_requirements, new_subprocess = True)

        #Cycle over the jobs and store the info
        for sj in jobSlice:
            #Check we are able to get the job status - if not set to failed.
            if sj.backend.id not in statusList['Value'].keys():
                logger.error("Job %s with DIRAC ID %s has been removed from DIRAC. Unable to finalise it." % (sj.getFQID(), sj.backend.id))
                sj.force_status('failed')
                continue
            #If we wanted the sandbox make sure it downloaded OK.
            if downloadSandbox and not returnDict[sj.backend.id]['outSandbox']['OK']:
                logger.error("Output sandbox error for job %s: %s. Unable to finalise it." % (sj.getFQID(), returnDict[sj.backend.id]['outSandbox']['Message']))
                sj.force_status('failed')
                continue
            #Set the CPU time
            sj.backend.normCPUTime = returnDict[sj.backend.id]['cpuTime']

            # Set DiracFile metadata - this circuitous route is copied from the standard so as to be consistent with the rest of the job finalisation code.
            if hasattr(sj.outputfiles, 'get'):
                wildcards = [f.namePattern for f in sj.outputfiles.get(DiracFile) if regex.search(f.namePattern) is not None]
            else:
                wildcards = []

            lfn_store = os.path.join(sj.getOutputWorkspace().getPath(), getConfig('Output')['PostProcessLocationsFileName'])

            # Make the file on disk with a nullop...
            if not os.path.isfile(lfn_store):
                with open(lfn_store, 'w'):
                    pass

            if hasattr(sj.outputfiles, 'get') and sj.outputfiles.get(DiracFile):
                # Now we can iterate over the contents of the file without touching it
                with open(lfn_store, 'ab') as postprocesslocationsfile:
                    if not hasattr(returnDict[sj.backend.id]['outDataInfo'], 'keys'):
                        logger.error("Error understanding OutputDataInfo: %s" % str(returnDict[sj.backend.id]['outDataInfo']))
                        raise GangaDiracError("Error understanding OutputDataInfo: %s" % str(returnDict[sj.backend.id]['outDataInfo']))

                    ## Caution is not clear atm whether this 'Value' is an LHCbism or bug
                    list_of_files = returnDict[sj.backend.id]['outDataInfo'].get('Value', returnDict[sj.backend.id]['outDataInfo'].keys())
                    for file_name in list_of_files:
                        file_name = os.path.basename(file_name)
                        info = returnDict[sj.backend.id]['outDataInfo'].get(file_name)
                        #logger.debug("file_name: %s,\tinfo: %s" % (str(file_name), str(info)))

                        if not hasattr(info, 'get'):
                            logger.error("Error getting OutputDataInfo for: %s" % str(sj.getFQID('.')))
                            logger.error("Please check the Dirac Job still exists or attempt a job.backend.reset() to try again!")
                            logger.error("Err: %s" % str(info))
                            logger.error("file_info_dict: %s" % str(returnDict[sj.backend.id]['outDataInfo']))
                            raise GangaDiracError("Error getting OutputDataInfo")

                        valid_wildcards = [wc for wc in wildcards if fnmatch.fnmatch(file_name, wc)]
                        if not valid_wildcards:
                            valid_wildcards.append('')

                        for wc in valid_wildcards:
                            #logger.debug("wildcard: %s" % str(wc))

                            DiracFileData = 'DiracFile:::%s&&%s->%s:::%s:::%s\n' % (wc,
                                                                                    file_name,
                                                                                    info.get('LFN', 'Error Getting LFN!'),
                                                                                    str(info.get('LOCATIONS', ['NotAvailable'])),
                                                                                    info.get('GUID', 'NotAvailable')
                                                                                    )
                            #logger.debug("DiracFileData: %s" % str(DiracFileData))
                            postprocesslocationsfile.write(DiracFileData)
                            postprocesslocationsfile.flush()

                logger.debug("Written: %s" % open(lfn_store, 'r').readlines())

            #Set the status of the subjob
            sj.updateStatus(statusmapping[statusList['Value'][sj.backend.id]['Status']])

    @staticmethod
    def requeue_dirac_finished_jobs(requeue_jobs, finalised_statuses):
        """
        Method used to requeue jobs whih are in the finalized state of some form, finished/failed/etc
        Args:
            requeue_jobs (list): This is a list of the jobs which are to be requeued to be finalised
            finalised_statuses (dict): Dict of the Dirac statuses vs the Ganga statuses after running
        """

        # requeue existing completed job
        for j in requeue_jobs:
            if j.been_queued:
                continue

            if monitoring_component:
                if monitoring_component.should_stop():
                    break
            # Job has changed underneath us don't attempt to finalize
            if j.backend.status not in finalised_statuses:
                j.been_queued = False
                continue
            if not configDirac['serializeBackend']:
                getQueues()._monitoring_threadpool.add_function(DiracBase.job_finalisation,
                                                           args=(j, finalised_statuses[j.backend.status]),
                                                           priority=5, name="Job %s Finalizing" % j.fqid)
                j.been_queued = True
            else:
                DiracBase.job_finalisation(j, finalised_statuses[j.backend.status])


    @staticmethod
    def monitor_dirac_running_jobs(monitor_jobs, finalised_statuses):
        """
        Method to update the configuration of jobs which are in a submitted/running state in Ganga&Dirac
        Args:
            monitor_jobs (list): Jobs which are to be monitored for their status change
            finalised_statuses (dict): Dict of the Dirac statuses vs the Ganga statuses after running
        """

        # now that can submit in non_blocking mode, can see jobs in submitting
        # that have yet to be assigned an id so ignore them
        # NOT SURE THIS IS VALID NOW BULK SUBMISSION IS GONE
        # EVEN THOUGH COULD ADD queues.add(j.submit) WILL KEEP AN EYE ON IT
        # dirac_job_ids    = [ j.backend.id for j in monitor_jobs if j.backend.id is not None ]
        # Correction this did become a problem for a crashed session during
        # submit, see #104454
        dead_jobs = (j for j in monitor_jobs if j.backend.id is None)
        for d in dead_jobs:
            d.updateStatus('failed')
            if d.master is not None:
                d.master.updateMasterJobStatus()

        ganga_job_status = [j.status for j in monitor_jobs if j.backend.id is not None]
        dirac_job_ids = [j.backend.id for j in monitor_jobs if j.backend.id is not None]

        logger.debug("GangaStatus: %s" % str(ganga_job_status))
        logger.debug("diracJobIDs: %s" % str(dirac_job_ids))

        if not dirac_job_ids:
            ## Nothing to do here stop bugging DIRAC about it!
            ## Everything else beyond here in the function depends on some ids present here, no ids means we can stop.
            return

        statusmapping = configDirac['statusmapping']

        result, bulk_state_result = execute('monitorJobs(%s, %s)' %( repr(dirac_job_ids), repr(statusmapping)), cred_req=monitor_jobs[0].backend.credential_requirements, new_subprocess = True)

        #result = results[0]
        #bulk_state_result = results[1]

        if len(result) != len(ganga_job_status):
            logger.warning('Dirac monitoring failed for %s, result = %s' % (str(dirac_job_ids), str(result)))
            logger.warning("Results: %s" % str(result))
            return

        requeue_job_list = []
        jobStateDict = {}

        jobs_to_update = {}
        master_jobs_to_update = []

        thread_handled_states = ['completed', 'failed']
        for job, state, old_state in zip(monitor_jobs, result, ganga_job_status):
            if monitoring_component and monitoring_component.should_stop():
                    break

            if job.been_queued:
                continue

            job.backend.statusInfo = state[0]
            job.backend.status = state[1]
            job.backend.actualCE = state[2]
            updated_dirac_status = state[3]
            try:
                job.backend.extraInfo = state[4]
            except Exception as err:
                logger.debug("gexception: %s" % str(err))
                pass
            logger.debug('Job status vector  : ' + job.fqid + ' : ' + repr(state))

            if updated_dirac_status not in jobStateDict:
                jobStateDict[updated_dirac_status] = []
            jobStateDict[updated_dirac_status].append(job)

            if job.backend.status in finalised_statuses:
                if job.status != 'running':
                    if job.status in ['completing', 'completed']:
                        continue
                    elif job.status in ['removed', 'killed']:
                        requeue_job_list.append(job)
                    elif (job.master and job.master.status in ['removed', 'killed']):
                        continue  # user changed it under us
                    else:
                        if 'running' not in jobs_to_update:
                            jobs_to_update['running'] = []
                        jobs_to_update['running'].append(job)
                        if job.master:
                            if job.master not in master_jobs_to_update:
                                master_jobs_to_update.append(job.master)
                        requeue_job_list.append(job)

            else:
                if job.status in ['removed', 'killed']:
                    continue
                elif (job.master and job.master.status in ['removed', 'killed']):
                    continue  # user changed it under us
                elif job.status != updated_dirac_status:
                    if job.status in ['completing', 'completed']:
                        # Another thread got there first
                        continue
                    if updated_dirac_status not in jobs_to_update:
                        jobs_to_update[updated_dirac_status] = []
                    jobs_to_update[updated_dirac_status].append(job)
                    if job.master:
                        if job.master not in master_jobs_to_update:
                            master_jobs_to_update.append(job.master)

        DiracBase._bulk_updateStateTime(jobStateDict, bulk_state_result)

        for status in jobs_to_update:
            for job in jobs_to_update[status]:
                job.updateStatus(status, update_master=False)

        for j in master_jobs_to_update:
            j.updateMasterJobStatus()

        DiracBase.requeue_dirac_finished_jobs(requeue_job_list, finalised_statuses)

    @staticmethod
    def updateMonitoringInformation(jobs_):
        """Check the status of jobs and retrieve output sandboxesi
        Args:
            jobs_ (list): List of the appropriate jobs to monitored
        """
        # Only those jobs in 'submitted','running' are passed in here for checking
        # if however they have already completed in Dirac they may have been put on queue
        # for processing from last time. These should be put back on queue without
        # querying dirac again. Their signature is status = running and job.backend.status
        # already set to Done or Failed etc.

        jobs = [stripProxy(j) for j in jobs_]

        # remove from consideration any jobs already in the queue. Checking this non persisted attribute
        # is better than querying the queue as cant tell if a job has just been taken off queue and is being processed
        # also by not being persistent, this attribute automatically allows queued jobs from last session to be considered
        # for requeing
        interesting_jobs = [j for j in jobs if not j.been_queued]
        # status that correspond to a ganga 'completed' or 'failed' (see DiracCommands.status(id))
        # if backend status is these then the job should be on the queue
        finalised_statuses = configDirac['finalised_statuses']

        monitor_jobs = [j for j in interesting_jobs if j.backend.status not in finalised_statuses]
        requeue_jobs = [j for j in interesting_jobs if j.backend.status in finalised_statuses]

        #logger.debug('Interesting jobs: ' + repr([j.fqid for j in interesting_jobs]))
        #logger.debug('Monitor jobs    : ' + repr([j.fqid for j in monitor_jobs]))
        #logger.debug('Requeue jobs    : ' + repr([j.fqid for j in requeue_jobs]))

        try:
            # Split all the monitorable jobs into groups based on the
            # credential used to communicate with DIRAC
            for requeue_jobs_group in group_jobs_by_backend_credential(requeue_jobs):
                DiracBase.requeue_dirac_finished_jobs(requeue_jobs_group, finalised_statuses)
            for monitor_jobs_group in group_jobs_by_backend_credential(monitor_jobs):
                DiracBase.monitor_dirac_running_jobs(monitor_jobs_group, finalised_statuses)
        except GangaDiracError as err:
            logger.warning("Error in Monitoring Loop, jobs on the DIRAC backend may not update")
            logger.debug(err)

def finalise_jobs_func(jobs, getSandbox = True):
    """Finalise the provided set of jobs."""
    DiracBase.finalise_jobs(jobs, getSandbox)

exportToGPI('finalise_jobs', finalise_jobs_func, 'Functions')


#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#

